# -*- coding: utf-8 -*-
"""
TODO:
    * cross validation
    * encounter vs database (time filtering)
"""
import logging
import functools
import copy
import utool as ut
import numpy as np
import six
from wbia.control import controller_inject

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

VERB_TESTDATA = ut.get_verbflag('testdata', 'td', 'acfg')[0]

SEED1 = 0
SEED2 = 42

if False and ut.is_developer():
    USE_ACFG_CACHE = (
        not ut.get_argflag(('--nocache-annot', '--nocache-aid', '--nocache'))
        and ut.USE_CACHE
    )
    USE_ACFG_CACHE = False
else:
    USE_ACFG_CACHE = False

_tup = controller_inject.make_ibs_register_decorator(__name__)
CLASS_INJECT_KEY, register_ibs_method = _tup


@profile
def time_filter_annots():
    r"""
    python -m wbia.init.filter_annots time_filter_annots \
            --db PZ_Master1 -a ctrl:qmingt=2 --profile

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> result = time_filter_annots()
    """
    import wbia

    wbia.testdata_expanded_aids()


@register_ibs_method
def filter_annots_general(ibs, aid_list=None, filter_kw={}, verbose=False, **kwargs):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        aid_list (list):  list of annotation rowids
        filter_kw (?):

    KWargs::
        has_none_annotmatch, any_match_annotmatch, has_all, is_known,
        any_match_annot, logic_annot, none_match_annotmatch,
        max_num_annotmatch, any_startswith_annot, has_any, require_quality,
        species, any_match, view_ext, has_any_annotmatch, view_pername,
        max_num_annot, min_timedelta, any_startswith, max_numfeat,
        any_startswith_annotmatch, been_adjusted, any_endswith_annot,
        require_viewpoint, logic, has_any_annot, min_num_annotmatch, min_num,
        min_num_annot, has_all_annot, has_none, min_pername,
        any_endswith_annotmatch, any_endswith, require_timestamp, none_match,
        contributor_contains, has_all_annotmatch, logic_annotmatch, min_numfeat,
        none_match_annot, view_ext1, view_ext2, max_num, has_none_annot,
        minqual, view

    CommandLine:
        python -m wbia --tf filter_annots_general
        python -m wbia --tf filter_annots_general --db PZ_Master1 \
                --has_any=[needswork,correctable,mildviewpoint] \
                --has_none=[viewpoint,photobomb,error:viewpoint,quality] --show

        python -m wbia --tf filter_annots_general --db=GZ_Master1  \
                --max-numfeat=300 --show --minqual=junk --species=None
        python -m wbia --tf filter_annots_general --db=lynx \
                --been_adjusted=True

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> import wbia
        >>> filter_kw = ut.argparse_dict(get_default_annot_filter_form(),
        >>>                              type_hint=ut.ddict(list, has_any=list,
        >>>                                                 has_none=list,
        >>>                                                 logic=str))
        >>> print('filter_kw = %s' % (ut.repr2(filter_kw),))
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aid_list = ibs.get_valid_aids()
        >>> #filter_kw = dict(is_known=True, min_num=1, has_any='viewpoint')
        >>> #filter_kw = dict(is_known=True, min_num=1, any_match='.*error.*')
        >>> aid_list_ = filter_annots_general(ibs, aid_list, filter_kw)
        >>> print('len(aid_list_) = %r' % (len(aid_list_),))
        >>> all_tags = ut.flatten(ibs.get_annot_all_tags(aid_list_))
        >>> filtered_tag_hist = ut.dict_hist(all_tags)
        >>> ut.print_dict(filtered_tag_hist, key_order_metric='val')
        >>> ut.print_dict(ibs.get_annot_stats_dict(aid_list_), 'annot_stats')
        >>> ut.quit_if_noshow()
        >>> import wbia.viz.interact
        >>> wbia.viz.interact.interact_chip.interact_multichips(ibs, aid_list_)
        >>> ut.show_if_requested()
    """
    if aid_list is None:
        aid_list = ibs.get_valid_aids()
    filter_kw_ = get_default_annot_filter_form()
    ut.update_existing(filter_kw_, filter_kw, iswarning=True, assert_exists=True)
    ut.update_existing(filter_kw_, kwargs, iswarning=True, assert_exists=True)
    aid_list_ = aid_list
    # filter_kw = ut.merge_dicts(get_default_annot_filter_form(), filter_kw)
    # TODO MERGE FILTERFLAGS BY TAGS AND FILTERFLAGS INDEPENDANT
    # aid_list_ = ibs.filterannots_by_tags(aid_list_, filter_kw)
    aid_list_ = ibs.filter_annots_independent(aid_list_, filter_kw_, verbose=verbose)
    aid_list_ = filter_annots_intragroup(ibs, aid_list_, filter_kw_, verbose=verbose)
    return aid_list_


@register_ibs_method
def sample_annots_general(ibs, aid_list=None, filter_kw={}, verbose=False, **kwargs):
    """ filter + sampling """
    # hack
    from wbia.expt import annotation_configs

    if aid_list is None:
        aid_list = ibs.get_valid_aids()
    filter_kw_ = annotation_configs.INDEPENDENT_DEFAULTS.copy()
    filter_kw_.update(annotation_configs.SUBINDEX_DEFAULTS.copy())
    filter_kw_.update(annotation_configs.SAMPLE_DEFAULTS.copy())

    ut.update_existing(filter_kw_, filter_kw, iswarning=True, assert_exists=True)
    ut.update_existing(filter_kw_, kwargs, iswarning=True, assert_exists=True)
    aid_list_ = aid_list
    # filter_kw = ut.merge_dicts(get_default_annot_filter_form(), filter_kw)
    # TODO MERGE FILTERFLAGS BY TAGS AND FILTERFLAGS INDEPENDANT
    # aid_list_ = ibs.filterannots_by_tags(aid_list_, filter_kw)
    aid_list_ = ibs.filter_annots_independent(aid_list_, filter_kw_, verbose=verbose)
    aid_list_ = filter_annots_intragroup(ibs, aid_list_, filter_kw_, verbose=verbose)

    aid_list_ = sample_annots(ibs, aid_list_, filter_kw_, verbose=verbose)
    aid_list_ = subindex_annots(ibs, aid_list_, filter_kw_, verbose=verbose)
    return aid_list_


@profile
def get_default_annot_filter_form():
    r"""
    Returns dictionary containing defaults for all valid filter parameters

    CommandLine:
        python -m wbia --tf get_default_annot_filter_form

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> filter_kw = get_default_annot_filter_form()
        >>> print(ut.repr2(filter_kw, align=True))
        >>> print(', '.join(filter_kw.keys()))
    """
    from wbia.expt import annotation_configs

    iden_defaults = annotation_configs.INDEPENDENT_DEFAULTS.copy()
    filter_kw = iden_defaults
    # tag_defaults = get_annot_tag_filterflags(
    #    None, None, {}, request_defaultkw=True)
    # filter_kw = ut.dict_union3(iden_defaults, tag_defaults, combine_op=None)
    return filter_kw


@register_ibs_method
def get_annot_tag_filterflags(ibs, aid_list, filter_kw, request_defaultkw=False):
    r"""
    Filters annotations by tags including those that is belongs to in a pair
    """
    from wbia import tag_funcs

    # Build Filters
    filter_keys = ut.get_func_kwargs(tag_funcs.filterflags_general_tags)

    annotmatch_filterkw = {}
    annot_filterkw = {}
    both_filterkw = {}

    kwreg = ut.KWReg(enabled=request_defaultkw)

    for key in filter_keys:
        annotmatch_filterkw[key] = filter_kw.get(*kwreg(key + '_annotmatch', None))
        annot_filterkw[key] = filter_kw.get(*kwreg(key + '_annot', None))
        both_filterkw[key] = filter_kw.get(*kwreg(key, None))

    if request_defaultkw:
        return kwreg.defaultkw

    # Grab Data
    need_annot_tags = any([var is not None for var in annot_filterkw.values()])
    need_annotmatch_tags = any([var is not None for var in annotmatch_filterkw.values()])
    need_both_tags = any([var is not None for var in both_filterkw.values()])

    if need_annot_tags or need_both_tags:
        annot_tags_list = ibs.get_annot_case_tags(aid_list)

    if need_annotmatch_tags or need_both_tags:
        annotmatch_tags_list = ibs.get_annot_annotmatch_tags(aid_list)

    if need_both_tags:
        both_tags_list = list(
            map(
                ut.unique_ordered,
                map(ut.flatten, zip(annot_tags_list, annotmatch_tags_list)),
            )
        )

    # Filter Data
    flags = np.ones(len(aid_list), dtype=np.bool)
    if need_annot_tags:
        flags_ = tag_funcs.filterflags_general_tags(annot_tags_list, **annot_filterkw)
        np.logical_and(flags_, flags, out=flags)

    if need_annotmatch_tags:
        flags_ = tag_funcs.filterflags_general_tags(
            annotmatch_tags_list, **annotmatch_filterkw
        )
        np.logical_and(flags_, flags, out=flags)

    if need_both_tags:
        flags_ = tag_funcs.filterflags_general_tags(both_tags_list, **both_filterkw)
        np.logical_and(flags_, flags, out=flags)
    return flags


@register_ibs_method
def filterannots_by_tags(ibs, aid_list, filter_kw):
    r"""
    Args:
        ibs (IBEISController):  wbia controller object
        aid_list (list):  list of annotation rowids

    CommandLine:
        python -m wbia --tf filterannots_by_tags
        utprof.py -m wbia --tf filterannots_by_tags

    SeeAlso:
        filter_annotmatch_by_tags

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='PZ_Master1')
        >>> aid_list = ibs.get_valid_aids()
        >>> has_any = ut.get_argval('--tags', type_=list,
        >>>                         default=['SceneryMatch', 'Photobomb'])
        >>> min_num = ut.get_argval('--min_num', type_=int, default=1)
        >>> filter_kw = dict(has_any=has_any, min_num=1)
        >>> aid_list_ = filterannots_by_tags(ibs, aid_list, filter_kw)
        >>> print('aid_list_ = %r' % (aid_list_,))
        >>> ut.quit_if_noshow()
        >>> pass
        >>> # TODO: show special annot group in GUI
    """
    flags = get_annot_tag_filterflags(ibs, aid_list, filter_kw)
    aid_list_ = ut.compress(aid_list, flags)
    return aid_list_


def get_acfg_cacheinfo(ibs, aidcfg):
    """
    Returns location and name of the ~~annot~~ data cache
    """
    from os.path import dirname, join

    # Make loading aids a big faster for experiments
    if ut.is_developer():
        import wbia

        repodir = dirname(ut.get_module_dir(wbia))
        acfg_cachedir = join(repodir, 'ACFG_CACHE')
    else:
        # acfg_cachedir = './localdata/ACFG_CACHE'
        acfg_cachedir = join(ibs.get_cachedir(), 'ACFG_CACHE')
        ut.ensuredir(acfg_cachedir)
    acfg_cachename = 'ACFG_CACHE'

    RESPECT_INTERNAL_CFGS = False
    if RESPECT_INTERNAL_CFGS:
        aid_cachestr = ibs.get_dbname() + '_' + ut.hashstr27(ut.to_json(aidcfg))
    else:
        relevant_aidcfg = copy.deepcopy(aidcfg)
        ut.delete_dict_keys(relevant_aidcfg['qcfg'], ut.INTERNAL_CFGKEYS)
        ut.delete_dict_keys(relevant_aidcfg['dcfg'], ut.INTERNAL_CFGKEYS)
        aid_cachestr = ibs.get_dbname() + '_' + ut.hashstr27(ut.to_json(relevant_aidcfg))
    acfg_cacheinfo = (acfg_cachedir, acfg_cachename, aid_cachestr)
    return acfg_cacheinfo


@profile
def expand_single_acfg(ibs, aidcfg, verbose=None):
    """
    for main_helpers"""
    from wbia.expt import annotation_configs

    if verbose is None:
        verbose = VERB_TESTDATA
    if verbose:
        logger.info('+=== EXPAND_SINGLE_ACFG ===')
        logger.info(
            ' * acfg = %s'
            % (ut.repr2(annotation_configs.compress_aidcfg(aidcfg), align=True),)
        )
        logger.info('+---------------------')
    avail_aids = ibs._get_all_aids()
    avail_aids = ibs.filter_annotation_set(avail_aids, is_staged=False)
    avail_aids = filter_annots_independent(ibs, avail_aids, aidcfg, verbose=verbose)
    avail_aids = filter_annots_intragroup(ibs, avail_aids, aidcfg, verbose=verbose)
    avail_aids = sample_annots(ibs, avail_aids, aidcfg, verbose=verbose)
    avail_aids = subindex_annots(ibs, avail_aids, aidcfg, verbose=verbose)
    aids = avail_aids
    if verbose:
        logger.info('L___ EXPAND_SINGLE_ACFG ___')
    return aids


@profile
def hack_remove_label_errors(ibs, expanded_aids, verbose=None):
    qaids_, daids_ = expanded_aids

    partitioned_sets = ibs.partition_annots_into_corresponding_groups(qaids_, daids_)
    tup = partitioned_sets
    query_group, data_group, unknown_group, distract_group = tup

    unknown_flags = ibs.unflat_map(
        ibs.get_annot_tag_filterflags,
        unknown_group,
        filter_kw=dict(none_match=['.*error.*']),
    )
    # data_flags  = ibs.unflat_map(
    #    ibs.get_annot_tag_filterflags, data_group,
    #    filter_kw=dict(none_match=['.*error.*']))
    query_flags = ibs.unflat_map(
        ibs.get_annot_tag_filterflags,
        query_group,
        filter_kw=dict(none_match=['.*error.*']),
    )

    query_noterror_flags = list(
        map(
            all,
            ut.list_zipflatten(
                query_flags,
                # data_flags,
            ),
        )
    )
    unknown_noterror_flags = list(map(all, unknown_flags))

    filtered_queries = ut.flatten(ut.compress(query_group, query_noterror_flags))
    filtered_unknown = ut.flatten(ut.compress(unknown_group, unknown_noterror_flags))

    filtered_qaids_ = sorted(filtered_queries + filtered_unknown)

    expanded_aids = (filtered_qaids_, daids_)

    if verbose:
        ut.colorprint('+---------------------', 'red')
        ibs.print_annotconfig_stats(filtered_qaids_, daids_)
        ut.colorprint('L___ HACKED_EXPAND_ACFGS ___', 'red')
    return expanded_aids


@profile
def hack_extra(ibs, expanded_aids):
    # SUCH HACK to get a larger database
    from wbia.expt import annotation_configs

    _aidcfg = annotation_configs.default['dcfg']
    _aidcfg['sample_per_name'] = 1
    _aidcfg['sample_size'] = 500
    _aidcfg['min_pername'] = 1
    _aidcfg['require_viewpoint'] = True
    _aidcfg['exclude_reference'] = True
    _aidcfg['view'] = 'right'
    prefix = 'hack'
    qaids = expanded_aids[0]
    daids = expanded_aids[1]

    _extra_aids = ibs.get_valid_aids()
    _extra_aids = ibs.remove_groundtrue_aids(_extra_aids, (qaids + daids))
    _extra_aids = filter_annots_independent(ibs, _extra_aids, _aidcfg, prefix)
    _extra_aids = sample_annots(ibs, _extra_aids, _aidcfg, prefix)
    daids = sorted(daids + _extra_aids)
    expanded_aids = (qaids, daids)
    return expanded_aids


def expand_acfgs_consistently(
    ibs, acfg_combo, initial_aids=None, use_cache=None, verbose=None, base=0
):
    r"""
    Expands a set of configurations such that they are comparable

    CommandLine:
        python -m wbia --tf parse_acfg_combo_list  \
                -a varysize
        wbia --tf get_annotcfg_list --db PZ_Master1 -a varysize
        #wbia --tf get_annotcfg_list --db lynx -a default:hack_imageset=True
        wbia --tf get_annotcfg_list --db PZ_Master1 -a varysize:qsize=None
        wbia --tf get_annotcfg_list --db PZ_Master0 --nofilter-dups  -a varysize
        wbia --tf get_annotcfg_list --db PZ_MTEST -a varysize --nofilter-dups
        wbia --tf get_annotcfg_list --db PZ_Master0 --verbtd \
                --nofilter-dups -a varysize
        wbia --tf get_annotcfg_list --db PZ_Master1 -a viewpoint_compare \
                --verbtd --nofilter-dups
        wbia --tf get_annotcfg_list -a timectrl --db GZ_Master1 --verbtd \
                --nofilter-dups

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> from wbia.init import main_helpers
        >>> from wbia.expt import annotation_configs
        >>> from wbia.expt.experiment_helpers import parse_acfg_combo_list
        >>> import wbia
        >>> ibs = wbia.opendb('PZ_MTEST')
        >>> #acfg_name_list = ['timectrl:dpername=[1,2]']
        >>> acfg_name_list = ['default:crossval_enc=True,require_timestamp=True']
        >>> aids = ibs.get_valid_aids()
        >>> main_helpers.monkeypatch_encounters(ibs, aids, days=50)
        >>> acfg_combo_list = parse_acfg_combo_list(acfg_name_list)
        >>> acfg_combo = acfg_combo_list[0]
        >>> initial_aids = None
        >>> use_cache = False
        >>> verbose = False
        >>> expanded_aids_combo_list = expand_acfgs_consistently(
        >>>     ibs, acfg_combo, initial_aids=initial_aids, use_cache=use_cache,
        >>>     verbose=verbose)
        >>> # Restore state
        >>> main_helpers.unmonkeypatch_encounters(ibs)
        >>> ut.assert_eq(len(expanded_aids_combo_list), 5)
    """
    from wbia.expt import annotation_configs
    import copy

    if verbose is None:
        verbose = VERB_TESTDATA
    # Edit configs so the sample sizes are consistent
    # FIXME: requiers that smallest configs are specified first

    def tmpmin(a, b):
        if a is None:
            return b
        elif b is None:
            return a
        return min(a, b)

    expanded_aids_list = []

    # Keep track of seen samples
    min_qsize = None
    min_dsize = None

    # HACK: Find out the params being varied and disallow those from being
    # prefiltered due to the lack of heirarchical filters
    nonvaried_dict, varied_acfg_list = annotation_configs.partition_acfg_list(acfg_combo)
    hack_exclude_keys = list(
        set(
            ut.flatten(
                [list(ut.merge_dicts(*acfg.values()).keys()) for acfg in varied_acfg_list]
            )
        )
    )

    # HACK: determine unconstrained min / max nannots
    acfg_combo_in = copy.deepcopy(acfg_combo)

    if False:
        acfg_combo2 = copy.deepcopy(acfg_combo_in)

        unconstrained_expansions = []
        for combox, acfg in enumerate(acfg_combo2):
            qcfg = acfg['qcfg']
            dcfg = acfg['dcfg']
            with ut.Indenter('[PRE %d] ' % (combox,)):
                expanded_aids = expand_acfgs(
                    ibs,
                    acfg,
                    initial_aids=initial_aids,
                    use_cache=use_cache,
                    hack_exclude_keys=hack_exclude_keys,
                    verbose=verbose,
                )
                unconstrained_expansions.append(expanded_aids)

        if any(ut.take_column(ut.take_column(acfg_combo_in, 'dcfg'), 'force_const_size')):
            unconstrained_lens = np.array(
                [(len(q), len(d)) for q, d in unconstrained_expansions]
            )
            # max_dlen = unconstrained_lens.T[1].max()
            min_dlen = unconstrained_lens.T[1].min()

            for acfg in acfg_combo_in:
                dcfg = acfg['dcfg']
                # TODO: make sample size annot_sample_size
                # sample size is #annots
                if dcfg['sample_size'] is None:
                    dcfg['_orig_sample_size'] = dcfg['sample_size']
                    dcfg['sample_size'] = min_dlen

    acfg_combo_out = []

    for combox, acfg in enumerate(acfg_combo_in):
        qcfg = acfg['qcfg']
        dcfg = acfg['dcfg']

        # In some cases we may want to clamp these, but others we do not
        if qcfg['force_const_size']:
            qcfg['_orig_sample_size'] = qcfg['sample_size']
            qcfg['sample_size'] = tmpmin(qcfg['sample_size'], min_qsize)

        if dcfg['force_const_size']:
            dcfg['_orig_sample_size'] = dcfg['sample_size']
            dcfg['sample_size'] = tmpmin(dcfg['sample_size'], min_dsize)

        # Expand modified acfgdict
        with ut.Indenter('[%d] ' % (combox,)):
            expanded_aids = expand_acfgs(
                ibs,
                acfg,
                initial_aids=initial_aids,
                use_cache=use_cache,
                hack_exclude_keys=hack_exclude_keys,
                verbose=verbose,
            )

            # if dcfg.get('hack_extra', None):
            #    assert False
            #    expanded_aids = hack_extra(ibs, expanded_aids)

            qsize = len(expanded_aids[0])
            dsize = len(expanded_aids[1])

            # <hack for float that should not interfere with other hacks
            if qcfg['sample_size'] != qsize:
                qcfg['_orig_sample_size'] = qcfg['sample_size']
            if dcfg['sample_size'] != dsize:
                dcfg['_orig_sample_size'] = dcfg['sample_size']
            # /-->

            if min_qsize is None:
                qcfg['sample_size'] = qsize
            if min_dsize is None:  # UNSURE
                dcfg['sample_size'] = dsize

            if qcfg['sample_size'] != qsize:
                qcfg['_true_sample_size'] = qsize
            if dcfg['sample_size'] != dsize:
                dcfg['_true_sample_size'] = dsize

            if qcfg['force_const_size']:
                min_qsize = tmpmin(min_qsize, qsize)
            if dcfg['force_const_size']:  # UNSURE
                min_dsize = tmpmin(min_dsize, dsize)

            # so hacky
            # this has to be after sample_size assignment, otherwise the filtering
            # is unstable Remove queries that have labeling errors in them.
            # TODO: fix errors AND remove labels
            # remove_label_errors = ut.is_developer() or ut.get_argflag('--noerrors')
            # ut.is_developer() or ut.get_argflag('--noerrors')
            remove_label_errors = qcfg.get('hackerrors', False)
            if remove_label_errors:
                expanded_aids = hack_remove_label_errors(ibs, expanded_aids, verbose)

        if qcfg['crossval_enc'] or dcfg['crossval_enc']:
            # Hack, just use qaids for cross validated sampleing
            aids = expanded_aids[0]
            qenc_per_name = qcfg['crossval_enc']
            denc_per_name = dcfg['crossval_enc']
            if qenc_per_name is None:
                qenc_per_name = 1
            if denc_per_name is None:
                denc_per_name = 1
            crossval_expansion = encounter_crossval(
                ibs, aids, qenc_per_name=qenc_per_name, denc_per_name=denc_per_name
            )

            import uuid

            unique_joinme = uuid.uuid4()

            for count, aid_pairs in enumerate(crossval_expansion):
                acfg_out = copy.deepcopy(acfg)
                acfg_out['qcfg']['crossval_idx'] = count
                acfg_out['dcfg']['crossval_idx'] = count
                acfg_out['qcfg']['sample_size'] = len(aid_pairs[0])
                acfg_out['dcfg']['sample_size'] = len(aid_pairs[1])
                # FIMXE: needs to be different for all acfgs
                # out of this sample.
                if acfg_out['qcfg'].get('joinme', None) is None:
                    acfg_out['qcfg']['joinme'] = unique_joinme
                    acfg_out['dcfg']['joinme'] = unique_joinme
                # need further hacks to assign sample size correctly
                # after the crossval hack
                acfg_combo_out.append(acfg_out)
            expanded_aids_list.extend(crossval_expansion)

        else:
            acfg_combo_out.append(acfg)
            # ibs.print_annotconfig_stats(*expanded_aids)
            expanded_aids_list.append(expanded_aids)

    # Sample afterwords
    return list(zip(acfg_combo_out, expanded_aids_list))


def crossval_helper(
    nid_to_sample_pool,
    perquery,
    perdatab,
    n_need,
    n_splits=None,
    rng=None,
    rebalance=True,
):
    """
    does sampling based on some grouping (or no grouping) of annots

    perquery = 2
    perdatab = 2

    nid_to_sample_pool = {
        1: [1, 2, 3, 4],
        2: [6, 7, 8, 9],
    }
    """
    if len(nid_to_sample_pool) == 0:
        raise ValueError(
            'Names do not have enough data for %d/%d split' % (perquery, perdatab)
        )
    rng = ut.ensure_rng(rng, impl='python')

    def split_combos(pool, perquery, perdatab, rng):
        import scipy.special

        poolsize = len(pool)
        # Number of ways we can select queries
        n_qmax = int(scipy.special.comb(poolsize, perquery))
        # Number of ways we can select targets from remaining items
        n_dmax = int(scipy.special.comb(poolsize - perquery, perdatab))
        # Total number of query / data combinations
        n_combos = n_qmax * n_dmax

        # Yield random combinations until we get something we havent seen
        poolset = set(pool)
        splits = set()
        while len(splits) < n_combos:
            # combo = tuple(sorted(rng.choice(items, size, replace=False)))
            qcombo = tuple(sorted(rng.sample(pool, perquery)))
            remain = poolset - set(qcombo)
            dcombo = tuple(sorted(rng.sample(remain, perdatab)))
            # TODO: try not to use queries / databases that we've used before
            # until we've exhauseted those possibilities.
            split = (qcombo, dcombo)
            if split not in splits:
                splits.add(split)
                yield split

    if n_splits is None:
        # What is the maximum number of items in a name?
        maxsize_name = max(map(len, nid_to_sample_pool.values()))
        # This is only a heuristic
        n_splits = maxsize_name

    # Create a mapping from each name to a list of query/target splits
    nid_to_splits = ut.ddict(list)
    # Create several splits for each name
    for nid, pool in nid_to_sample_pool.items():
        # Randomly select up to `n_splits` combinations of size `n_need`.
        combo_iter = split_combos(pool, perquery, perdatab, rng)
        for count, fold_split in enumerate(combo_iter, start=1):
            # Earlier samples will be biased towards names with more annots
            nid_to_splits[nid].append(fold_split)
            if count >= n_splits:
                break

    # logger.info(ut.repr2(list(nid_to_splits.values()), strvals=True, nl=2))

    # Some names may have more splits than others
    # nid_to_nsplits = ut.map_vals(len, nid_to_splits)
    # Find the name with the most splits
    # max_nid = ut.argmax(nid_to_nsplits)
    # max_size = nid_to_nsplits[max_nid]

    new_splits = [[] for _ in range(n_splits)]
    if rebalance:
        # Rebalance by adding combos from each name in a cycle.
        # The difference between the largest and smallest split is at most one.
        for count, split in enumerate(ut.iflatten(nid_to_splits.values())):
            new_splits[count % len(new_splits)].append(split)
    else:
        # No rebalancing. The first split contains everything from the dataset
        # and subsequent splits contain less and less.
        for nid, combos in nid_to_splits.items():
            for count, split in enumerate(combos):
                new_splits[count].append(split)

    # Reshape into an expanded aids list
    # List of query / database objects per split, grouped by name.
    reshaped_splits = [ut.listT(splits) for splits in new_splits if len(splits) > 0]
    return reshaped_splits


def encounter_crossval(
    ibs,
    aids,
    qenc_per_name=1,
    denc_per_name=1,
    enc_labels=None,
    confusors=True,
    rng=None,
    annots_per_enc=None,
    rebalance=True,
    n_splits=None,
    early=False,
):
    r"""
    Constructs a list of [ (qaids, daids) ] where there are `qenc_per_name` and
    `denc_per_name` for each individual in the datasets respectively.
    `enc_labels` specifies custom encounter labels.

    CommandLine:
        python -m wbia.init.filter_annots encounter_crossval

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> from wbia.init import main_helpers
        >>> import wbia
        >>> #ibs, aids = wbia.testdata_aids(
        >>> #    defaultdb='WWF_Lynx_Copy',
        >>> #    a='default:minqual=good,require_timestamp=True,view=left')
        >>> ibs, aids = wbia.testdata_aids(defaultdb='PZ_MTEST',
        >>>                                 a='default:require_timestamp=True')
        >>> main_helpers.monkeypatch_encounters(ibs, aids, days=50)
        >>> qenc_per_name = 2
        >>> denc_per_name = 2
        >>> confusors = False
        >>> print('denc_per_name = %r' % (denc_per_name,))
        >>> print('qenc_per_name = %r' % (qenc_per_name,))
        >>> rng = 0
        >>> n_splits = 5
        >>> expanded_aids = encounter_crossval(ibs, aids, n_splits=n_splits,
        >>>                                    qenc_per_name=qenc_per_name,
        >>>                                    denc_per_name=denc_per_name,
        >>>                                    confusors=confusors, rng=rng)
        >>> # ensure stats agree
        >>> cfgargs = dict(per_vp=False, per_multiple=False, combo_dists=False,
        >>>                per_name=False, per_enc=True, use_hist=False)
        >>> for qaids, daids in expanded_aids:
        >>>     stats = ibs.get_annotconfig_stats(qaids, daids, **cfgargs)
        >>>     del stats['confusor_daid_stats']
        >>>     print(ut.repr2(stats, strvals=True, strkeys=True, nl=2))
        >>>     denc_stats = stats['matchable_daid_stats']['denc_per_name']
        >>>     qenc_stats = stats['qaid_stats']['qenc_per_name']
        >>>     assert denc_stats['min'] == denc_stats['max']
        >>>     assert denc_stats['min'] == denc_per_name
        >>>     assert qenc_stats['min'] == qenc_stats['max']
        >>>     assert qenc_stats['min'] == qenc_per_name
        >>> # Restore state
        >>> main_helpers.unmonkeypatch_encounters(ibs)
        >>> #qaids, daids = expanded_aids[0]
        >>> #stats = ibs.get_annotconfig_stats(qaids, daids, use_hist=True)
        >>> #print(ut.repr2(stats, strvals=True, strkeys=True, nl=2))

    """
    qenc_per_name = int(qenc_per_name)
    denc_per_name = int(denc_per_name)

    # We can only select individuals with enough encounters
    # Any name without enought data becomes a confusor
    n_need = qenc_per_name + denc_per_name
    perquery = qenc_per_name
    perdatab = denc_per_name

    annots = ibs.annots(aids)
    if enc_labels is None:
        enc_labels = annots.encounter_text

    # Group annotations by encounter
    encounters = ibs._annot_groups(annots.group(enc_labels)[1])
    enc_nids = ut.take_column(encounters.nids, 0)
    rng = ut.ensure_rng(rng, impl='python')
    if annots_per_enc is not None:
        encounters = [rng.sample(list(a), annots_per_enc) for a in encounters]
    # Group encounters by name
    nid_to_encs = ut.group_items(encounters, enc_nids)

    nid_to_confusors = {
        nid: ut.lmap(tuple, encs)
        for nid, encs in nid_to_encs.items()
        if len(encs) < n_need
    }
    nid_to_sample_pool = {
        nid: ut.lmap(tuple, encs)
        for nid, encs in nid_to_encs.items()
        if len(encs) >= n_need
    }

    reshaped_splits = crossval_helper(
        nid_to_sample_pool,
        perquery,
        perdatab,
        n_splits=n_splits,
        n_need=n_need,
        rng=rng,
        rebalance=rebalance,
    )
    if early:
        return reshaped_splits, nid_to_confusors

    # logger.info(ut.repr4(reshaped_splits, nl=3))
    expanded_aids_list = [
        [ut.flatten(qpart), ut.flatten(dpart)] for qpart, dpart in reshaped_splits
    ]

    expanded_aids_list = [
        [sorted(ut.flatten(ut.flatten(qpart))), sorted(ut.flatten(ut.flatten(dpart)))]
        for qpart, dpart in reshaped_splits
    ]

    if confusors:
        # Add confusors the the dataset
        confusor_aids = ut.flatten(ut.flatten(nid_to_confusors.values()))
        expanded_aids_list = [
            (qaids, sorted(daids + confusor_aids)) for qaids, daids in expanded_aids_list
        ]
    return expanded_aids_list


def annot_crossval(
    ibs,
    aid_list,
    n_qaids_per_name=1,
    n_daids_per_name=1,
    rng=None,
    debug=True,
    n_splits=None,
    confusors=True,
):
    """
    Stratified sampling per name size

    Args:
        n_splits (int): number of query/database splits to create.
            note, some names may not be big enough to split this many times.

    CommandLine:
        python -m wbia.init.filter_annots annot_crossval

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> aid_list = ibs.get_valid_aids()
        >>> n_qaids_per_name = 2
        >>> n_daids_per_name = 3
        >>> rng = 0
        >>> debug = True
        >>> n_splits = None
        >>> expanded_aids_list = annot_crossval(
        >>>     ibs, aid_list, n_qaids_per_name, n_daids_per_name, rng, debug,
        >>>     n_splits, confusors=False)
        >>> result = ('expanded_aids_list = %s' % (ut.repr2(expanded_aids_list, nl=2),))
        >>> print(result)
    """
    # Parameters
    n_need = n_qaids_per_name + n_daids_per_name
    rebalance = True

    # Group annotations by name
    annots = ibs.annots(aids=aid_list)
    nid_to_aids = ut.group_items(annots.aids, annots.nids)

    # Any name without enough data becomes a confusor
    # Otherwise we can use it in the sampling pool
    nid_to_confusors = {
        nid: aids for nid, aids in nid_to_aids.items() if len(aids) < n_need
    }
    nid_to_sample_pool = {
        nid: aids for nid, aids in nid_to_aids.items() if len(aids) >= n_need
    }

    perquery = n_qaids_per_name
    perdatab = n_daids_per_name

    reshaped_splits = crossval_helper(
        nid_to_sample_pool,
        perquery,
        perdatab,
        n_splits=n_splits,
        n_need=n_need,
        rng=rng,
        rebalance=rebalance,
    )

    expanded_aids_list = [
        [sorted(ut.flatten(qpart)), sorted(ut.flatten(dpart))]
        for qpart, dpart in reshaped_splits
    ]

    if confusors:
        # Add confusors the the dataset
        confusor_aids = ut.flatten(nid_to_confusors.values())
        expanded_aids_list = [
            (qaids, sorted(daids + confusor_aids)) for qaids, daids in expanded_aids_list
        ]

    # if debug:
    #     debug_expanded_aids(expanded_aids_list)
    return expanded_aids_list

    # # Group annotations by encounter
    # unique_encounters, groupxs = annots.group_indicies(enc_labels)
    # encounter_nids = annots.take(ut.take_column(groupxs, 0)).nid
    # encounter_aids = ut.apply_grouping(annots.aids, groupxs)
    # # Group encounters by name
    # nid_to_encs = ut.group_items(encounter_aids, encounter_nids)

    # nid_to_confusors = {nid: enc for nid, enc in nid_to_encs.items()
    #                     if len(enc) <  n_need}
    # nid_to_sample_pool = {nid: enc for nid, enc in nid_to_encs.items()
    #                       if len(enc) >= n_need}

    # # Randomly shuffle encounters for each individual
    # valid_enc_rand_idxs = [(nid, ut.random_indexes(len(enc), rng=rng))
    #                        for nid, enc in nid_to_sample_pool.items()]

    # # For each individual choose a set of query and database encounters
    # crossval_idx_samples = []
    # max_num_encounters = max(map(len, (t[1] for t in valid_enc_rand_idxs)))
    # # TODO: iterate over a sliding window to choose multiple queries OR
    # # iterate over all combinations of queries... (harder)
    # for i in range(max_num_encounters):
    #     encx_split = {}
    #     for nid, idxs in valid_enc_rand_idxs:
    #         if i < len(idxs):
    #             # Choose the database encounters from anything not chosen as a
    #             # query
    #             d_choices = ut.where(ut.not_list(
    #                 ut.index_to_boolmask([i], len(idxs))))
    #             js = rng.choice(d_choices, size=denc_per_name, replace=False)
    #             encx_split[nid] = (idxs[i:i + 1], idxs[js])
    #     crossval_idx_samples.append(encx_split)

    # # convert to aids
    # crossval_aid_samples = []
    # for encx_split in crossval_idx_samples:
    #     aid_split = {}
    #     for nid, (qxs, dxs) in encx_split.items():
    #         qaids = ut.flatten(ut.take(nid_to_sample_pool[nid], qxs))
    #         daids = ut.flatten(ut.take(nid_to_sample_pool[nid], dxs))
    #         assert len(dxs) == denc_per_name
    #         assert len(qxs) == qenc_per_name
    #         aid_split[nid] = (qaids, daids)
    #     crossval_aid_samples.append(aid_split)

    # tups = [(nid, aids_) for aid_split_ in crossval_aid_samples
    #         for nid, aids_ in aid_split_.items()]
    # groups = ut.take_column(tups, 0)
    # aidpairs = ut.take_column(tups, 1)
    # # crossval_samples[0]

    # # rebalance the queries
    # # Rewrite using rebalance code from crossval_annots
    # # Very inefficient but does what I want
    # group_to_idxs = ut.dzip(*ut.group_indices(groups))
    # freq = ut.dict_hist(groups)
    # g = list(freq.keys())[ut.argmax(list(freq.values()))]
    # size = freq[g]
    # new_splits = [[] for _ in range(size)]
    # while True:
    #     try:
    #         g = list(freq.keys())[ut.argmax(list(freq.values()))]
    #         if freq[g] == 0:
    #             raise StopIteration()
    #         group_idxs = group_to_idxs[g]
    #         group_to_idxs[g] = []
    #         freq[g] = 0
    #         priorityx = ut.argsort(list(map(len, new_splits)))
    #         for nextidx, splitx in zip(group_idxs, priorityx):
    #             new_splits[splitx].append(nextidx)
    #     except StopIteration:
    #         break
    # # name_splits = ut.unflat_take(groups, new_splits)
    # aid_splits = ut.unflat_take(aidpairs, new_splits)

    # # Add annots that could not meet these requirements are distractors
    # confusors = ut.flatten(ut.flatten(list(nid_to_confusors.values())))

    # expanded_aids_list = []
    # for aidsplit in aid_splits:
    #     qaids = sorted(ut.flatten(ut.take_column(aidsplit, 0)))
    #     daids = sorted(ut.flatten(ut.take_column(aidsplit, 1)) + confusors)
    #     expanded_aids_list.append((qaids, daids))
    # return expanded_aids_list


@profile
def expand_acfgs(
    ibs,
    aidcfg,
    verbose=None,
    use_cache=None,
    hack_exclude_keys=None,
    initial_aids=None,
    save_cache=True,
):
    r"""
    Main multi-expansion function. Expands an annot config dict into qaids and
    daids.  New version of this function based on a configuration dictionary
    built from command line argumetns

    Args:
        ibs (IBEISController):  wbia controller object
        aidcfg (dict): configuration of the annotation filter
        verbose (bool):  verbosity flag(default = False)
        use_cache (bool):  turns on disk based caching(default = None)
        hack_exclude_keys (None): (default = None)
        initial_aids (None): (default = None)

    Returns:
        tuple: expanded_aids=(qaid_list, daid_list) - expanded list of aids
            that meet the criteria of the aidcfg filter

    TODO:
        The database should be created first in most circumstances, then
        the queries should be filtered to meet the database restrictions?
        I'm not sure Sometimes you need to set the query aids constant, but
        sometimes you need to set the data aids constant. Seems to depend.

        This function very much needs the idea of filter chains

        OkNewIdea:
            3 filters:
                * Common sampling - takes care of things like min time delta,
                * species, quality viewpoint etc.
                * query sampling
                * database sampling
            Basic idea is
                * Sample large pool
                * Partition pool into query and database
            Requires:
                * base sampling params
                * partition1 params
                * partition2 params
                * inter partition params?

    CommandLine:
        python -m wbia.dev -e print_acfg  -a timectrl:qsize=10,dsize=10  --db PZ_MTEST --veryverbtd --nocache-aid
        python -m wbia.dev -e print_acfg  -a timectrl:qminqual=good,qsize=10,dsize=10  --db PZ_MTEST --veryverbtd --nocache-aid

        python -m wbia.dev -e print_acfg  -a timectrl --db PZ_MTEST --verbtd --nocache-aid
        python -m wbia.dev -e print_acfg  -a timectrl --db PZ_Master1 --verbtd --nocache-aid
        python -m wbia.dev -e print_acfg  -a timequalctrl --db PZ_Master1 --verbtd --nocache-aid

        python -m wbia.dev -e rank_cmc   -a controlled:qsize=10,dsize=10,dper_name=2 -t default --db PZ_MTEST
        python -m wbia.dev -e rank_cmc   -a controlled:qsize=10,dsize=20,dper_name=2 -t default --db PZ_MTEST
        python -m wbia.dev -e print      -a controlled:qsize=10,dsize=10             -t default --db PZ_MTEST --verbtd --nocache-aid

        python -m wbia.dev -e latexsum -t candinvar -a viewpoint_compare  --db NNP_Master3 --acfginfo
        utprof.py -m wbia.dev -e print -t candk -a varysize  --db PZ_MTEST --acfginfo
        utprof.py -m wbia.dev -e latexsum -t candk -a controlled  --db PZ_Master0 --acfginfo

        python -m wbia --tf get_annotcfg_list:0 --db NNP_Master3 -a viewpoint_compare --nocache-aid --verbtd

        python -m wbia --tf get_annotcfg_list  --db PZ_Master1 \
            -a timectrl:qhas_any=\(needswork,correctable,mildviewpoint\),qhas_none=\(viewpoint,photobomb,error:viewpoint,quality\) \
            --acfginfo --veryverbtd  --veryverbtd
        python -m wbia --tf draw_rank_cmc --db PZ_Master1 --show -t best \
            -a timectrl:qhas_any=\(needswork,correctable,mildviewpoint\),qhas_none=\(viewpoint,photobomb,error:viewpoint,quality\) \
            --acfginfo --veryverbtd

        python -m wbia --tf get_annotcfg_list  --db Oxford -a default:qhas_any=\(query,\),dpername=2,exclude_reference=True --acfginfo --verbtd  --veryverbtd --nocache-aid

    CommandLine:
        python -m wbia.init.filter_annots --exec-expand_acfgs --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> import wbia
        >>> from wbia.expt import annotation_configs
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> aidcfg = copy.deepcopy(annotation_configs.default)
        >>> aidcfg['qcfg']['species'] = 'primary'
        >>> initial_aids = None
        >>> expanded_aids = expand_acfgs(ibs, aidcfg, initial_aids=initial_aids)
        >>> result = ut.repr3(expanded_aids, nl=1, nobr=True)
        >>> print(result)
        [1, 2, 3, 4, 5, 6],
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13],
    """
    from wbia.expt import annotation_configs

    if verbose is None:
        verbose = VERB_TESTDATA

    assert isinstance(aidcfg, dict), 'type(aidcfg)=%r' % (type(aidcfg),)
    aidcfg = copy.deepcopy(aidcfg)

    # Check if this filter has been cached
    # TODO: keep a database state config that augments the cachestr?
    if use_cache is None:
        use_cache = USE_ACFG_CACHE

    # save_cache = True
    if use_cache and save_cache:
        acfg_cacheinfo = get_acfg_cacheinfo(ibs, aidcfg)
        acfg_cachedir, acfg_cachename, aid_cachestr = acfg_cacheinfo
    if use_cache:
        aids_tup = ut.tryload_cache(acfg_cachedir, acfg_cachename, aid_cachestr)
        if aids_tup is not None:
            (qaid_list, daid_list) = aids_tup
            return qaid_list, daid_list

    comp_acfg = annotation_configs.compress_aidcfg(aidcfg)

    if verbose:
        ut.colorprint('+=== EXPAND_ACFGS ===', 'yellow')
        logger.info(' * acfg = %s' % (ut.repr2(comp_acfg, align=True),))
        ut.colorprint('+---------------------', 'yellow')

    # Breakup into common, query, and database configs
    qcfg = aidcfg['qcfg']
    dcfg = aidcfg['dcfg']
    common_cfg = comp_acfg['common']

    # Extract the common independent filtering params
    idenfilt_cfg_default = annotation_configs.INDEPENDENT_DEFAULTS
    idenfilt_cfg_empty = {key: None for key in idenfilt_cfg_default.keys()}
    idenfilt_cfg_common = ut.update_existing(idenfilt_cfg_empty, common_cfg, copy=True)

    if hack_exclude_keys:
        for key in hack_exclude_keys:
            if key in idenfilt_cfg_common:
                idenfilt_cfg_common[key] = None

    # Find the q/d specific filtering flags that were already taken care of in
    # common filtering. Set them all to None, so we dont rerun that filter
    qpredone_iden_keys = ut.dict_isect(qcfg, idenfilt_cfg_common).keys()
    for key in qpredone_iden_keys:
        qcfg[key] = None

    dpredone_iden_keys = ut.dict_isect(dcfg, idenfilt_cfg_common).keys()
    for key in dpredone_iden_keys:
        dcfg[key] = None

    # if aidcfg['qcfg']['hack_imageset'] is True:
    #    return ibs.get_imageset_expanded_aids()
    # Hack: Make hierarchical filters to supersede this
    if initial_aids is None:
        initial_aids = ibs._get_all_aids()

    initial_aids = ibs.filter_annotation_set(initial_aids, is_staged=False)

    verbflags = dict(verbose=verbose)
    qfiltflags = dict(prefix='q', **verbflags)
    dfiltflags = dict(prefix='d', **verbflags)

    default_aids = initial_aids

    # A chain of filters on all of the aids
    global_filter_chain = [
        (filter_annots_independent, idenfilt_cfg_common),
        (filter_annots_intragroup, idenfilt_cfg_common),
    ]

    # Chains of filters individually for each partition
    partition_chains = [
        [
            # Query partition chain
            (filter_annots_independent, qcfg),
            (filter_annots_intragroup, qcfg),
            (sample_annots, qcfg),
        ],
        [
            # Database partition chain
            (filter_annots_independent, dcfg),
            (filter_annots_intragroup, dcfg),
            (sample_annots_wrt_ref, dcfg, 0),
        ],
    ]
    try:
        # GLOBAL FILTER CHAIN
        # applies filtering to all available aids
        for filtfn, filtcfg in global_filter_chain:
            default_aids = filtfn(
                ibs, default_aids, filtcfg, prefix='', withpre=True, **verbflags
            )

        # PARTITION FILTER CHAIN
        # chain of filters for query / database annots
        default_qaids = default_daids = default_aids
        partition_avail_aids = [default_qaids, default_daids]
        partion_kwargs = [qfiltflags, dfiltflags]
        for index in range(len(partition_chains)):
            filter_chain = partition_chains[index]
            avail_aids = partition_avail_aids[index]
            _partkw = partion_kwargs[index].copy()
            for filter_tup in filter_chain:
                filtfn, filtcfg = filter_tup[0:2]
                if len(filter_tup) == 3:
                    # handle filters that take reference sets
                    refindex = filter_tup[2]
                    ref_aids = partition_avail_aids[refindex]
                    _partkw['ref_aids'] = ref_aids
                # Execute filtering
                avail_aids = filtfn(ibs, avail_aids, filtcfg, **_partkw)
            partition_avail_aids[index] = avail_aids

        # SUBINDEX EACH PARTITIONED CHAIN
        subindex_cfgs = [qcfg, dcfg]
        for index in range(len(partition_avail_aids)):
            avail_aids = partition_avail_aids[index]
            _partkw = partion_kwargs[index]
            filtcfg = subindex_cfgs[index]
            avail_aids = subindex_annots(ibs, avail_aids, filtcfg, **_partkw)
            partition_avail_aids[index] = avail_aids

        # UNPACK FILTER RESULTS
        avail_qaids, avail_daids = partition_avail_aids

    except Exception as ex:
        logger.info('PRINTING ERROR INFO')
        logger.info(' * acfg = %s' % (ut.repr2(comp_acfg, align=True),))
        ut.printex(ex, 'Error executing filter chains')
        raise

    qaid_list = sorted(avail_qaids)
    daid_list = sorted(avail_daids)

    if verbose:
        ut.colorprint('+---------------------', 'yellow')
        ibs.print_annotconfig_stats(qaid_list, daid_list)
        ut.colorprint('L___ EXPAND_ACFGS ___', 'yellow')

    # Save filter to cache
    if use_cache and save_cache:
        ut.ensuredir(acfg_cachedir)
        try:
            ut.save_cache(
                acfg_cachedir, acfg_cachename, aid_cachestr, (qaid_list, daid_list)
            )
        except IOError:
            pass
    return qaid_list, daid_list


def expand_species(ibs, species, avail_aids=None):
    if species == 'primary':
        species = ibs.get_primary_database_species()
    if species is None and avail_aids is not None:
        species = ibs.get_dominant_species(avail_aids)
    return species


@profile
@register_ibs_method
def filter_annots_independent(
    ibs, avail_aids, aidcfg, prefix='', verbose=VERB_TESTDATA, withpre=False
):
    r"""
    Filtering that doesn't have to do with a reference set of aids

    TODO make filterflags version

    Args:
        ibs (IBEISController):  wbia controller object
        avail_aids (list):
        aidcfg (dict):
        prefix (str): (default = '')
        verbose (bool):  verbosity flag(default = False)

    Returns:
        list: avail_aids

    CommandLine:
        python -m wbia --tf filter_annots_independent --veryverbtd

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> import wbia
        >>> from wbia.expt import annotation_configs
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> avail_aids = input_aids = ibs.get_valid_aids()
        >>> aidcfg = annotation_configs.default['dcfg']
        >>> aidcfg['require_timestamp'] = True
        >>> aidcfg['require_quality'] = False
        >>> aidcfg['is_known'] = True
        >>> prefix = ''
        >>> verbose = True
        >>> avail_aids = filter_annots_independent(ibs, avail_aids, aidcfg,
        >>>                                        prefix, verbose)
        >>> result = ('avail_aids = %s' % (str(avail_aids),))
        >>> print(result)

    Ignore:
        # Testing tag features
        python -m wbia --tf draw_rank_cmc --db PZ_Master1 --show -t best \
            -a timectrl:qhas_any=\(needswork,correctable,mildviewpoint\),qhas_none=\(viewpoint,photobomb,error:viewpoint,quality\) \
            ---acfginfo --veryverbtd
    """
    from wbia.other import ibsfuncs

    if aidcfg is None:
        if verbose:
            logger.info('No annot filter returning')
        return avail_aids

    VerbosityContext = verb_context('FILTER_INDEPENDENT', aidcfg, verbose)
    VerbosityContext.startfilter(withpre=withpre)

    if aidcfg.get('is_known') is True:
        with VerbosityContext('is_known'):
            avail_aids = ibs.filter_aids_without_name(
                avail_aids, invert=not aidcfg['is_known']
            )
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('is_exemplar') is not None:
        flags = ibs.get_annot_exemplar_flags(avail_aids)
        is_valid = [flag == aidcfg['is_exemplar'] for flag in flags]
        with VerbosityContext('is_exemplar'):
            avail_aids = ut.compress(avail_aids, is_valid)
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('reviewed') is not None:
        flags = ibs.get_annot_reviewed(avail_aids)
        is_valid = [flag == aidcfg['reviewed'] for flag in flags]
        with VerbosityContext('reviewed'):
            avail_aids = ut.compress(avail_aids, is_valid)
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('multiple') is not None:
        flags = ibs.get_annot_multiple(avail_aids)
        is_valid = [flag == aidcfg['multiple'] for flag in flags]
        with VerbosityContext('multiple'):
            avail_aids = ut.compress(avail_aids, is_valid)
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('require_timestamp') is True:
        with VerbosityContext('require_timestamp'):
            avail_aids = ibs.filter_aids_without_timestamps(avail_aids)

    if aidcfg.get('require_gps') is True:
        with VerbosityContext('require_gps'):
            annots = ibs.annots(avail_aids)
            annots = annots.compress(~np.isnan(np.array(annots.gps)).any(axis=1))
            avail_aids = annots.aids
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('max_timestamp') is not None:
        with VerbosityContext('max_timestamp'):
            max_dt = aidcfg.get('max_timestamp')
            import datetime

            if isinstance(max_dt, (int, float)):
                max_unixtime = max_dt
            elif isinstance(max_dt, six.string_types):
                if max_dt == 'now':
                    max_dt = datetime.datetime.utcnow()
                else:
                    max_dt = max_dt.replace('/', '-')
                    y, m, d = max_dt.split('-')
                    max_dt = ut.date_to_datetime(datetime.date(y, m, d))
            max_unixtime = ut.datetime_to_posixtime(max_dt)
            unixtimes = np.array(ibs.annots(avail_aids).image_unixtimes_asfloat)
            flag_list = np.logical_or(np.isnan(unixtimes), unixtimes <= max_unixtime)
            ut.compress(avail_aids, flag_list)
        # avail_aids = sorted(avail_aids)

    cfg_species = aidcfg.get('species')
    if isinstance(cfg_species, six.string_types) and cfg_species.lower() == 'none':
        cfg_species = None

    metadata = ut.LazyDict(species=lambda: expand_species(ibs, cfg_species, None))

    if cfg_species is not None:
        species = metadata['species']
        with VerbosityContext('species', species=species):
            avail_aids = ibs.filter_aids_to_species(avail_aids, species)
            # avail_aids = sorted(avail_aids)

    if aidcfg.get('been_adjusted', None):
        # HACK to see if the annotation has been adjusted from the default
        # value set by dbio.ingest_database
        flag_list = ibs.get_annot_been_adjusted(avail_aids)
        with VerbosityContext('been_adjusted'):
            avail_aids = ut.compress(avail_aids, flag_list)

    if aidcfg.get('contributor_contains', None):
        contributor_contains = aidcfg['contributor_contains']
        gid_list = ibs.get_annot_gids(avail_aids)
        tag_list = ibs.get_image_contributor_tag(gid_list)
        flag_list = [contributor_contains in tag for tag in tag_list]
        with VerbosityContext('contributor_contains'):
            avail_aids = ut.compress(avail_aids, flag_list)

    if aidcfg.get('minqual') is not None or aidcfg.get('require_quality'):
        minqual = 'junk' if aidcfg['minqual'] is None else aidcfg['minqual']
        with VerbosityContext('minqual', 'require_quality'):
            # Filter quality
            avail_aids = ibs.filter_aids_to_quality(
                avail_aids, minqual, unknown_ok=not aidcfg['require_quality']
            )
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('max_unixtime', None) is not None:
        max_unixtime = aidcfg.get('max_unixtime', None)
        unixtimes = np.array(ibs.get_annot_image_unixtimes_asfloat(avail_aids))
        flags = unixtimes <= max_unixtime
        with VerbosityContext('max_unixtime'):
            avail_aids = ut.compress(avail_aids, flags)
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('min_unixtime', None) is not None:
        min_unixtime = aidcfg.get('min_unixtime', None)
        unixtimes = np.array(ibs.get_annot_image_unixtimes_asfloat(avail_aids))
        flags = unixtimes >= min_unixtime
        with VerbosityContext('min_unixtime'):
            avail_aids = ut.compress(avail_aids, flags)
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('max_numfeat') is not None or aidcfg.get('min_numfeat') is not None:
        max_numfeat = aidcfg['max_numfeat']
        min_numfeat = aidcfg['min_numfeat']
        if max_numfeat is None:
            max_numfeat = np.inf
        if min_numfeat is None:
            min_numfeat = 0
        numfeat_list = np.array(ibs.get_annot_num_feats(avail_aids))
        flags_list = np.logical_and(
            numfeat_list >= min_numfeat, numfeat_list <= max_numfeat
        )
        with VerbosityContext('max_numfeat', 'min_numfeat'):
            avail_aids = ut.compress(avail_aids, flags_list)

    if aidcfg.get('view') is not None or aidcfg.get('require_viewpoint'):
        # Resolve base viewpoint
        if aidcfg['view'] == 'primary':
            view = ibsfuncs.get_primary_species_viewpoint(metadata['species'])
        elif aidcfg['view'] == 'primary1':
            view = ibsfuncs.get_primary_species_viewpoint(metadata['species'], 1)
        else:
            view = aidcfg['view']
        if isinstance(view, six.string_types) and view.lower() == 'none':
            view = None
        OLD = False
        if OLD:
            view_ext1 = (
                aidcfg['view_ext'] if aidcfg['view_ext1'] is None else aidcfg['view_ext1']
            )
            view_ext2 = (
                aidcfg['view_ext'] if aidcfg['view_ext2'] is None else aidcfg['view_ext2']
            )
            valid_yaws = ibsfuncs.get_extended_viewpoints(
                view, num1=view_ext1, num2=view_ext2
            )
            unknown_ok = not aidcfg['require_viewpoint']
            with VerbosityContext(
                'view',
                'require_viewpoint',
                'view_ext',
                'view_ext1',
                'view_ext2',
                valid_yaws=valid_yaws,
            ):
                avail_aids = ibs.filter_aids_to_viewpoint(
                    avail_aids, valid_yaws, unknown_ok=unknown_ok
                )
            avail_aids = sorted(avail_aids)
        else:

            def rectify_view(vstr):
                # FIXME: I stopped implementing the += stuff
                vstr_num = vstr.lower()
                num = 0
                if not vstr_num.endswith('1'):
                    vstr = vstr_num
                else:
                    if '+' in vstr:
                        vstr, numstr = vstr_num.split('+')
                        num = int(numstr)
                    if '-' in vstr:
                        vstr, numstr = vstr_num.split('+')
                        num = -int(numstr)
                assert num == 0, 'cant do += yet'
                if vstr == 'primary':
                    return ibsfuncs.get_primary_species_viewpoint(metadata['species'])
                for yawtxt, other_yawtxt in ibs.const.YAWALIAS.items():
                    other_yawtxt = ut.ensure_iterable(other_yawtxt)
                    if vstr == yawtxt.lower():
                        return yawtxt
                    for x in other_yawtxt:
                        if vstr == x.lower():
                            return yawtxt
                raise ValueError('unknown viewpoint vstr=%r' % (vstr,))

            if view is None:
                valid_yaw_txts = None
            else:
                valid_yaw_txts = [
                    rectify_view(vstr) for vstr in ut.smart_cast(view, list)
                ]
            unknown_ok = not aidcfg['require_viewpoint']
            yaw_flags = ibs.get_viewpoint_filterflags(
                avail_aids, valid_yaw_txts, unknown_ok=unknown_ok, assume_unique=True
            )
            yaw_flags = list(yaw_flags)
            with VerbosityContext(
                'view',
                'require_viewpoint',
                'view_ext',
                'view_ext1',
                'view_ext2',
                valid_yaws=valid_yaw_txts,
            ):
                avail_aids = ut.compress(avail_aids, yaw_flags)

    # if aidcfg.get('exclude_view') is not None:
    #    raise NotImplementedError('view tag resolution of exclude_view')
    #    # Filter viewpoint
    #    # TODO need to resolve viewpoints
    #    exclude_view = aidcfg.get('exclude_view')
    #    with VerbosityContext('exclude_view', hack=True):
    #        avail_aids = ibs.remove_aids_of_viewpoint(
    #            avail_aids, exclude_view)

    if aidcfg.get('min_pername_global') is not None:
        # Keep annots with at least this many groundtruths in the database
        min_pername_global = aidcfg.get('min_pername_global')
        num_gt_global_list = ibs.get_annot_num_groundtruth(avail_aids, noself=False)
        flag_list = np.array(num_gt_global_list) >= min_pername_global
        with VerbosityContext('exclude_view'):
            avail_aids = ut.compress(avail_aids, flag_list)
        # avail_aids = sorted(avail_aids)

    if aidcfg.get('max_pername_global') is not None:
        max_pername_global = aidcfg.get('max_pername_global')
        num_gt_global_list = ibs.get_annot_num_groundtruth(avail_aids, noself=False)
        flag_list = np.array(num_gt_global_list) <= max_pername_global
        with VerbosityContext('exclude_view'):
            avail_aids = ut.compress(avail_aids, flag_list)
        # avail_aids = sorted(avail_aids)

    # FILTER HACK integrating some notion of tag functions
    # TODO: further integrate
    if aidcfg.get('has_any', None) or aidcfg.get('has_none', None):
        filterkw = ut.dict_subset(aidcfg, ['has_any', 'has_none'], None)
        flags = get_annot_tag_filterflags(ibs, avail_aids, filterkw)
        with VerbosityContext('has_any', 'has_none'):
            avail_aids = ut.compress(avail_aids, flags)
        # avail_aids = sorted(avail_aids)

    avail_aids = sorted(avail_aids)

    VerbosityContext.endfilter()
    return avail_aids


@profile
def filter_annots_intragroup(
    ibs, avail_aids, aidcfg, prefix='', verbose=VERB_TESTDATA, withpre=False
):
    r"""
    This filters annots using information about the relationships
    between the annotations in the ``avail_aids`` group. This function is not
    independent and a second consecutive call may yield new results.
    Thus, the order in which this filter is applied matters.

    CommandLine:
        wbia --tf get_annotcfg_list \
                -a default:qsame_imageset=True,been_adjusted=True,excluderef=True \
                --db lynx --veryverbtd --nocache-aid

    Ignore:
        >>> aidcfg['min_timedelta'] = 60 * 60 * 24
        >>> aidcfg['min_pername'] = 3
    """
    from wbia.other import ibsfuncs

    if aidcfg is None:
        if verbose:
            logger.info('No annot filter returning')
        return avail_aids

    VerbosityContext = verb_context('FILTER_INTRAGROUP', aidcfg, verbose)
    VerbosityContext.startfilter(withpre=withpre)

    metadata = ut.LazyDict(
        species=lambda: expand_species(ibs, aidcfg['species'], avail_aids)
    )

    if aidcfg['same_imageset'] is not None:
        same_imageset = aidcfg['same_imageset']
        assert same_imageset is True
        imgsetid_list = ibs.get_annot_primary_imageset(avail_aids)
        nid_list = ibs.get_annot_nids(avail_aids)
        multiprop2_aids = ut.hierarchical_group_items(
            avail_aids, [nid_list, imgsetid_list]
        )
        qaid_list = []
        # TODO: sampling using different enouncters
        for imgsetid, nid2_aids in multiprop2_aids.items():
            if len(nid2_aids) == 1:
                pass
            else:
                aids_list = list(nid2_aids.values())
                idx = ut.list_argmax(list(map(len, aids_list)))
                qaids = aids_list[idx]
                qaid_list.extend(qaids)
        with VerbosityContext('same_imageset'):
            avail_aids = qaid_list
        avail_aids = sorted(avail_aids)

    # TODO:
    # Filter via GPS distance
    # try:
    #    if aidcfg['min_spacedelta'] is not None:
    #        pass
    #    if aidcfg['min_spacetimedelta'] is not None:
    #        pass
    # except KeyError:
    #    pass

    # FIXME: This is NOT an independent filter because it depends on pairwise
    # interactions
    if aidcfg['view_pername'] is not None:
        species = metadata['species']
        # This filter removes entire names.  The avaiable aids must be from
        # names with certain viewpoint frequency properties
        prop2_nid2_aids = ibs.group_annots_by_prop_and_name(
            avail_aids, ibs.get_annot_viewpoint_code
        )

        countstr = aidcfg['view_pername']
        primary_viewpoint = ibsfuncs.get_primary_species_viewpoint(species)
        lhs_dict = {
            'primary': primary_viewpoint,
            'primary1': ibsfuncs.get_extended_viewpoints(
                primary_viewpoint, num1=1, num2=0, include_base=False
            )[0],
        }
        self = ut.CountstrParser(lhs_dict, prop2_nid2_aids)
        nid2_flag = self.parse_countstr_expr(countstr)
        nid2_aids = ibs.group_annots_by_name_dict(avail_aids)
        valid_nids = [nid for nid, flag in nid2_flag.items() if flag]
        with VerbosityContext('view_pername', countstr=countstr):
            avail_aids = ut.flatten(ut.dict_take(nid2_aids, valid_nids))
        # avail_aids = sorted(avail_aids)

    if aidcfg['min_timedelta'] is not None:
        min_timedelta = ut.ensure_timedelta(aidcfg['min_timedelta'])
        with VerbosityContext('min_timedelta', min_timedelta=min_timedelta):
            avail_aids = ibs.filter_annots_using_minimum_timedelta(
                avail_aids, min_timedelta
            )
        # avail_aids = sorted(avail_aids)

    # Each aid must have at least this number of other groundtruth aids
    min_pername = aidcfg['min_pername']
    if min_pername is not None:
        grouped_aids_ = ibs.group_annots_by_name(
            avail_aids, distinguish_unknowns=True, assume_unique=True
        )[0]
        with VerbosityContext('min_pername'):
            flags = np.array(ut.lmap(len, grouped_aids_)) >= min_pername
            avail_aids = ut.flatten(ut.compress(grouped_aids_, flags))
            # avail_aids = ut.flatten([
            #    aids for aids in grouped_aids_ if len(aids) >= min_pername])
        # avail_aids = sorted(avail_aids)

    max_pername = aidcfg['max_pername']
    if max_pername is not None:
        grouped_aids_ = ibs.group_annots_by_name(
            avail_aids, distinguish_unknowns=True, assume_unique=True
        )[0]
        with VerbosityContext('max_pername'):
            avail_aids = ut.flatten(
                [aids for aids in grouped_aids_ if len(aids) <= max_pername]
            )
        # avail_aids = sorted(avail_aids)

    avail_aids = sorted(avail_aids)

    VerbosityContext.endfilter()
    return avail_aids


@profile
def get_reference_preference_order(
    ibs,
    gt_ref_grouped_aids,
    gt_avl_grouped_aids,
    prop_getter,
    cmp_func,
    aggfn,
    rng,
    verbose=VERB_TESTDATA,
):
    r"""
    Orders preference for sampling based on some metric
    """
    import vtool as vt

    grouped_reference_unixtimes = ibs.unflat_map(prop_getter, gt_ref_grouped_aids)
    grouped_available_gt_unixtimes = ibs.unflat_map(prop_getter, gt_avl_grouped_aids)

    grouped_reference_props = grouped_reference_unixtimes
    grouped_available_gt_props = grouped_available_gt_unixtimes

    # Order the available aids by some aggregation over some metric
    preference_scores = [
        aggfn(cmp_func(ref_prop, avl_prop[:, None]), axis=1)
        for ref_prop, avl_prop in zip(grouped_reference_props, grouped_available_gt_props)
    ]
    # Order by increasing timedelta (metric)
    gt_preference_idx_list = vt.argsort_groups(preference_scores, reverse=True, rng=rng)
    return gt_preference_idx_list


@profile
def sample_annots_wrt_ref(
    ibs, avail_aids, aidcfg, ref_aids, prefix='', verbose=VERB_TESTDATA
):
    """
    Sampling when a reference set is given
    """
    sample_per_name = aidcfg.get('sample_per_name')
    sample_per_ref_name = aidcfg.get('sample_per_ref_name')
    exclude_reference = aidcfg.get('exclude_reference')
    sample_size = aidcfg.get('sample_size')
    offset = aidcfg.get('sample_offset')
    sample_rule_ref = aidcfg.get('sample_rule_ref')
    sample_rule = aidcfg.get('sample_rule')
    sample_occur = aidcfg.get('sample_occur')
    exclude_ref_contact = aidcfg.get('exclude_ref_contact')

    avail_aids = sorted(avail_aids)
    ref_aids = sorted(ref_aids)

    VerbosityContext = verb_context('SAMPLE (REF)', aidcfg, verbose)
    VerbosityContext.startfilter()

    if sample_per_ref_name is None:
        sample_per_ref_name = sample_per_name

    if offset is None:
        offset = 0

    if exclude_reference:
        assert ref_aids is not None, 'ref_aids=%r' % (ref_aids,)
        # VerbosityContext.report_annot_stats(ibs, avail_aids, prefix, '')
        # VerbosityContext.report_annot_stats(ibs, ref_aids, prefix, '')
        with VerbosityContext('exclude_reference', num_ref_aids=len(ref_aids)):
            avail_aids = ut.setdiff_ordered(avail_aids, ref_aids)
            avail_aids = sorted(avail_aids)

    if exclude_ref_contact:
        with VerbosityContext('exclude_ref_contact', num_ref_aids=len(ref_aids)):
            # also_exclude_overlaps = ibs.get_dbname() == 'Oxford'
            contact_aids_list = ibs.get_annot_contact_aids(
                ref_aids, daid_list=avail_aids, assume_unique=True
            )
            # Disallow the same name in the same image
            x = ibs.unflat_map(ibs.get_annot_nids, contact_aids_list)
            y = ibs.get_annot_nids(ref_aids)
            sameimg_samename_aids = ut.flatten(
                [
                    ut.compress(aids, np.array(x0) == y0)
                    for aids, x0, y0 in zip(contact_aids_list, x, y)
                ]
            )
            # contact_aids = ut.flatten(contact_aids_list)
            avail_aids = ut.setdiff_ordered(avail_aids, sameimg_samename_aids)

    if sample_occur is True:
        with VerbosityContext('sample_occur', num_ref_aids=len(ref_aids)):
            # Get other aids from the references' encounters
            ref_enc_texts = ibs.get_annot_encounter_text(ref_aids)
            avail_enc_texts = ibs.get_annot_encounter_text(avail_aids)
            flags = ut.setdiff_flags(avail_enc_texts, ref_enc_texts)
            avail_aids = ut.compress(avail_aids, flags)

    if not (sample_per_ref_name is not None or sample_size is not None):
        VerbosityContext.endfilter()
        return avail_aids

    if ut.is_float(sample_size):
        # A float sample size is a interpolations between full data and small
        # data
        sample_size = int(
            round((len(avail_aids) * sample_size + (1 - sample_size) * len(ref_aids)))
        )
        if verbose:
            logger.info('Expanding sample size to: %r' % (sample_size,))

    # This function first partitions aids into a one set that corresonds with
    # the reference set and another that does not correspond with the reference
    # set. The rest of the filters operate on these sets independently
    partitioned_sets = ibs.partition_annots_into_corresponding_groups(
        ref_aids, avail_aids
    )
    # items
    # [0], and [1] are corresponding lists of annot groups
    # [2], and [3] are non-corresonding annot groups
    (
        gt_ref_grouped_aids,
        gt_avl_grouped_aids,
        gf_ref_grouped_aids,
        gf_avl_grouped_aids,
    ) = partitioned_sets

    if sample_per_ref_name is not None:
        rng = np.random.RandomState(SEED2)
        if sample_rule_ref == 'maxtimedelta':
            # Maximize time delta between query and corresponding database
            # annotations
            cmp_func = ut.absdiff
            aggfn = np.mean
            prop_getter = ibs.get_annot_image_unixtimes_asfloat
            gt_preference_idx_list = get_reference_preference_order(
                ibs,
                gt_ref_grouped_aids,
                gt_avl_grouped_aids,
                prop_getter,
                cmp_func,
                aggfn,
                rng,
            )
        elif sample_rule_ref == 'random':
            gt_preference_idx_list = [
                ut.random_indexes(len(aids), rng=rng) for aids in gt_avl_grouped_aids
            ]
        else:
            raise ValueError('Unknown sample_rule_ref = %r' % (sample_rule_ref,))
        gt_sample_idxs_list = ut.get_list_column_slice(
            gt_preference_idx_list, offset, offset + sample_per_ref_name
        )
        gt_sample_aids = ut.list_ziptake(gt_avl_grouped_aids, gt_sample_idxs_list)
        gt_avl_grouped_aids = gt_sample_aids

        with VerbosityContext(
            'sample_per_ref_name',
            'sample_rule_ref',
            'sample_offset',
            sample_per_ref_name=sample_per_ref_name,
        ):
            avail_aids = ut.flatten(gt_avl_grouped_aids) + ut.flatten(gf_avl_grouped_aids)

    if sample_per_name is not None:
        # sample rule is always random for gf right now
        rng = np.random.RandomState(SEED2)
        if sample_rule == 'random':
            gf_preference_idx_list = [
                ut.random_indexes(len(aids), rng=rng) for aids in gf_avl_grouped_aids
            ]
        else:
            raise ValueError('Unknown sample_rule=%r' % (sample_rule,))
        gf_sample_idxs_list = ut.get_list_column_slice(
            gf_preference_idx_list, offset, offset + sample_per_name
        )
        gf_sample_aids = ut.list_ziptake(gf_avl_grouped_aids, gf_sample_idxs_list)
        gf_avl_grouped_aids = gf_sample_aids

        with VerbosityContext('sample_per_name', 'sample_rule', 'sample_offset'):
            avail_aids = ut.flatten(gt_avl_grouped_aids) + ut.flatten(gf_avl_grouped_aids)

    gt_avl_aids = ut.flatten(gt_avl_grouped_aids)
    gf_avl_aids = ut.flatten(gf_avl_grouped_aids)

    if sample_size is not None:
        # Keep all correct matches to the reference set
        # We have the option of keeping ground false
        num_gt = len(gt_avl_aids)
        num_gf = len(gf_avl_aids)
        num_keep_gf = sample_size - num_gt
        num_remove_gf = num_gf - num_keep_gf
        if num_remove_gf < 0:
            # Too few ground false
            logger.info(
                (
                    'Warning: Cannot meet sample_size=%r. available_%saids '
                    'will be undersized by at least %d'
                )
                % (sample_size, prefix, -num_remove_gf)
            )
        if num_keep_gf < 0:
            # Too many multitons; Can never remove a multiton
            logger.info(
                'Warning: Cannot meet sample_size=%r. available_%saids '
                'will be oversized by at least %d' % (sample_size, prefix, -num_keep_gf)
            )
        rng = np.random.RandomState(SEED2)
        gf_avl_aids = ut.random_sample(gf_avl_aids, num_keep_gf, rng=rng)

        # random ordering makes for bad hashes
        with VerbosityContext(
            'sample_size',
            sample_size=sample_size,
            num_remove_gf=num_remove_gf,
            num_keep_gf=num_keep_gf,
        ):
            avail_aids = gt_avl_aids + gf_avl_aids

    avail_aids = sorted(gt_avl_aids + gf_avl_aids)

    VerbosityContext.endfilter()
    return avail_aids


@profile
def multi_sampled_seaturtle_queries():
    import wbia
    from wbia.expt import annotation_configs
    from wbia.expt import experiment_helpers
    from wbia.init.filter_annots import expand_acfgs
    import copy

    aidcfg = copy.deepcopy(annotation_configs.default)
    db = 'seaturtles'  # 'testdb1'
    ibs = wbia.opendb(defaultdb=db)
    a = [
        'default:sample_occur=True,occur_offset=0,exclude_reference=True,qhas_any=(left,right),num_names=1'
    ]
    acfg_combo_list = experiment_helpers.parse_acfg_combo_list(a)
    aidcfg = acfg_combo_list[0][0]

    if False:
        # Do each name individually. A bit slower, but more correct
        qaids_list = []
        daids_list = []
        aidcfg['qcfg']['name_offset'] = 0
        aidcfg['qcfg']['occur_offset'] = 0
        prev = -1
        while True:
            aidcfg['qcfg']['occur_offset'] = 0
            while True:
                qaids, daids = expand_acfgs(
                    ibs, aidcfg, use_cache=False, save_cache=False
                )
                aidcfg['qcfg']['occur_offset'] += 1
                if len(qaids) == 0:
                    break
                qaids_list.append(qaids)
                daids_list.append(daids)
                logger.info(qaids)
            if len(qaids_list) == prev:
                break
            prev = len(qaids_list)
            aidcfg['qcfg']['name_offset'] += 1

        for qaids, daids in zip(qaids_list, daids_list):
            ibs.print_annotconfig_stats(qaids, daids, enc_per_name=True, per_enc=True)
    else:
        # A bit faster because we can do multiple names at the same time
        qaids_list = []
        daids_list = []
        aidcfg['qcfg']['num_names'] = None
        aidcfg['dcfg']['num_names'] = None
        aidcfg['qcfg']['name_offset'] = 0
        aidcfg['qcfg']['occur_offset'] = 0
        while True:
            qaids, daids = expand_acfgs(ibs, aidcfg, use_cache=False, save_cache=False)
            aidcfg['qcfg']['occur_offset'] += 1
            if len(qaids) == 0:
                break
            qaids_list.append(qaids)
            daids_list.append(daids)
            logger.info(qaids)

        for qaids, daids in zip(qaids_list, daids_list):
            ibs.print_annotconfig_stats(qaids, daids, enc_per_name=True, per_enc=True)


@profile
def sample_annots(ibs, avail_aids, aidcfg, prefix='', verbose=VERB_TESTDATA):
    r"""
    Sampling preserves input sample structure and thust does not always return
    exact values

    CommandLine:
        python -m wbia --tf sample_annots --veryverbtd

        python -m wbia --tf get_annotcfg_list --db seaturtles \
            -a default:qhas_any=\(left,right\),sample_occur=True,exclude_reference=True,sample_offset=0,num_names=1 --acfginfo

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> import wbia
        >>> from wbia.expt import annotation_configs
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> avail_aids = input_aids = ibs.get_valid_aids()
        >>> aidcfg = copy.deepcopy(annotation_configs.default['dcfg'])
        >>> aidcfg['sample_per_name'] = 3
        >>> aidcfg['sample_size'] = 10
        >>> aidcfg['min_pername'] = 2
        >>> prefix = ''
        >>> verbose = True
        >>> avail_aids = filter_annots_independent(ibs, avail_aids, aidcfg,
        >>>                                        prefix, verbose)
        >>> avail_aids = sample_annots(ibs, avail_aids, aidcfg,
        >>>                            prefix, avail_aids)
        >>> result = ('avail_aids = %s' % (str(avail_aids),))
        >>> print(result)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.init.filter_annots import *  # NOQA
        >>> import wbia
        >>> from wbia.expt import annotation_configs
        >>> db = 'seaturtles'  # 'testdb1'
        >>> ibs = wbia.opendb(defaultdb=db)
        >>> aidcfg = copy.deepcopy(annotation_configs.default)['qcfg']
        >>> aidcfg['sample_occur'] = True
        >>> initial_aids = ibs.get_valid_aids()
        >>> withpre, verbose, prefix = True, 2, ''
        >>> avail_aids = filter_annots_independent(
        >>>     ibs, initial_aids, {'has_any': ['left', 'right']}, prefix, verbose)
        >>> qaids = sample_annots(ibs, avail_aids, aidcfg, prefix, verbose)
        >>> avail_aids = initial_aids
        >>> ref_aids = qaids
        >>> dcfg = dict(exclude_reference=True, sample_occur=True)
        >>> daids = sample_annots_wrt_ref(ibs, initial_aids, dcfg, qaids, prefix, verbose)
        >>> ibs.print_annotconfig_stats(qaids, daids, enc_per_name=True, per_enc=True)
    """
    import vtool as vt
    from wbia.expt import annotation_configs

    def get_cfg(key):
        default_dict = annotation_configs.SAMPLE_DEFAULTS
        return aidcfg.get(key, default_dict[key])

    VerbosityContext = verb_context('SAMPLE (NOREF)', aidcfg, verbose)
    VerbosityContext.startfilter()

    sample_rule = get_cfg('sample_rule')
    sample_per_name = get_cfg('sample_per_name')
    sample_size = get_cfg('sample_size')
    offset = get_cfg('sample_offset')
    occur_offset = get_cfg('occur_offset')
    name_offset = get_cfg('name_offset')
    num_names = get_cfg('num_names')
    sample_occur = get_cfg('sample_occur')

    unflat_get_annot_unixtimes = functools.partial(
        ibs.unflat_map, ibs.get_annot_image_unixtimes_asfloat
    )

    if offset is None:
        offset = 0
    if occur_offset is None:
        occur_offset = 0
    if name_offset is None:
        name_offset = 0

    if num_names is not None:
        grouped_aids = ibs.group_annots_by_name(avail_aids, assume_unique=True)[0]
        with VerbosityContext('num_names'):
            name_slice = slice(name_offset, name_offset + num_names)
            avail_aids = ut.flatten(grouped_aids[name_slice])

    if sample_occur is True:
        # Occurrence / Encounter sampling
        occur_texts = ibs.get_annot_occurrence_text(avail_aids)
        names = ibs.get_annot_names(avail_aids)
        grouped_ = ut.hierarchical_group_items(avail_aids, [names, occur_texts])
        # ensure dictionary ordering for offset consistency
        sgrouped_ = ut.sort_dict(ut.hmap_vals(ut.sort_dict, grouped_, max_depth=0))
        occur_slice = slice(occur_offset, occur_offset + 1)
        chosen = [
            ut.flatten(list(sub.values())[occur_slice]) for sub in sgrouped_.values()
        ]

        with VerbosityContext('sample_offset'):
            # TODO: num ocurrences to sample
            # TODO: num annots per encounter to sample
            avail_aids = ut.flatten(chosen)
        # now find which groups of annotations share those tags

    if sample_per_name is not None:
        # For the query we just choose a single annot per name
        # For the database we have to do something different
        grouped_aids = ibs.group_annots_by_name(avail_aids, assume_unique=True)[0]
        # Order based on some preference (like random)
        sample_seed = get_cfg('sample_seed')
        rng = np.random.RandomState(sample_seed)
        # + --- Get nested sample indicies ---
        if sample_rule == 'random':
            preference_idxs_list = [
                ut.random_indexes(len(aids), rng=rng) for aids in grouped_aids
            ]
        elif sample_rule == 'mintime':
            unixtime_list = unflat_get_annot_unixtimes(grouped_aids)
            preference_idxs_list = vt.argsort_groups(
                unixtime_list, reverse=False, rng=rng
            )
        elif sample_rule == 'maxtime':
            unixtime_list = unflat_get_annot_unixtimes(grouped_aids)
            preference_idxs_list = vt.argsort_groups(unixtime_list, reverse=True, rng=rng)
        elif sample_rule == 'qual_and_view':
            if sample_rule != 'qual_and_view':
                # Hacked in
                with VerbosityContext('sample_per_name', 'sample_rule', 'sample_offset'):
                    flags = ibs.get_annot_quality_viewpoint_subset(
                        avail_aids, annots_per_view=sample_per_name
                    )
                    avail_aids = ut.compress(avail_aids, flags)
        else:
            raise ValueError('Unknown sample_rule=%r' % (sample_rule,))
        # L ___
        if sample_rule != 'qual_and_view':
            sample_idxs_list = list(
                ut.iget_list_column_slice(
                    preference_idxs_list, offset, offset + sample_per_name
                )
            )
            sample_aids = ut.list_ziptake(grouped_aids, sample_idxs_list)

            with VerbosityContext('sample_per_name', 'sample_rule', 'sample_offset'):
                avail_aids = ut.flatten(sample_aids)
        avail_aids = sorted(avail_aids)

    if sample_size is not None:
        # BUG: Should sample annots while preserving name size
        if sample_size > len(avail_aids):
            logger.info('Warning sample size too large')
        rng = np.random.RandomState(SEED2)
        # Randomly sample names rather than annotations this makes sampling a
        # knapsack problem. Use a random greedy solution
        grouped_aids = ibs.group_annots_by_name(avail_aids, assume_unique=True)[0]
        # knapsack items values and weights are are num annots per name
        knapsack_items = [
            (len(aids), len(aids), count) for count, aids in enumerate(grouped_aids)
        ]
        ut.deterministic_shuffle(knapsack_items, rng=rng)
        total_value, items_subset = ut.knapsack_greedy(knapsack_items, sample_size)
        group_idx_sample = ut.get_list_column(items_subset, 2)
        subgroup_aids = ut.take(grouped_aids, group_idx_sample)
        with VerbosityContext('sample_size'):
            avail_aids = ut.flatten(subgroup_aids)
            # avail_aids = ut.random_sample(avail_aids, sample_size, rng=rng)
        if total_value != sample_size:
            logger.info('Sampling could not get exactly right sample size')
        avail_aids = sorted(avail_aids)

    VerbosityContext.endfilter()
    return avail_aids


@profile
def subindex_annots(
    ibs, avail_aids, aidcfg, ref_aids=None, prefix='', verbose=VERB_TESTDATA
):
    """
    Returns exact subindex of annotations
    """
    VerbosityContext = verb_context('SUBINDEX', aidcfg, verbose)
    VerbosityContext.startfilter(withpre=False)

    if aidcfg['shuffle']:
        rand_idx = ut.random_indexes(len(avail_aids), seed=SEED2)
        with VerbosityContext('shuffle', SEED2=SEED2):
            avail_aids = ut.take(avail_aids, rand_idx)

    if aidcfg['index'] is not None:
        indicies = ensure_flatlistlike(aidcfg['index'])
        _indexed_aids = [avail_aids[ix] for ix in indicies if ix < len(avail_aids)]
        with VerbosityContext('index', subset_size=len(_indexed_aids)):
            avail_aids = _indexed_aids

    # Always sort aids to preserve hashes? (Maybe sort the vuuids instead)
    avail_aids = sorted(avail_aids)

    VerbosityContext.endfilter(withpost=False)
    return avail_aids


@profile
def ensure_flatiterable(input_):
    if isinstance(input_, six.string_types):
        input_ = ut.fuzzy_int(input_)
    if isinstance(input_, int) or not ut.isiterable(input_):
        return [input_]
    elif isinstance(input_, (list, tuple)):
        # logger.info(input_)
        if len(input_) > 0 and ut.isiterable(input_[0]):
            return ut.flatten(input_)
        return input_
    else:
        raise TypeError('cannot ensure %r input_=%r is iterable', (type(input_), input_))


def ensure_flatlistlike(input_):
    # if isinstance(input_, slice):
    #    pass
    iter_ = ensure_flatiterable(input_)
    return list(iter_)


def verb_context(filtertype, aidcfg, verbose):
    """ closure helper """

    class VerbosityContext(object):
        """
        Printing filter info in a way that avoids polluting the function
        namespace. This is a hack.

        This is a with_statement context class that expect a variable avail_aids
        to be modified inside the context. It prints the state of the variable
        before and after filtering. Several static methods can be used
        at the start and end of larger filtering functions.
        """

        def __init__(self, *keys, **filterextra):
            self.prefix = ut.get_var_from_stack('prefix', verbose=False)
            if verbose:
                dictkw = dict(nl=False, explicit=True, nobraces=True)
                infostr = ''
                if len(keys) > 0:
                    subdict = ut.dict_subset(aidcfg, keys, None)
                    infostr += '' + ut.repr2(subdict, **dictkw)
                logger.info(
                    '[%s] * Filter by %s' % (self.prefix.upper(), infostr.strip())
                )
                if verbose > 1 and len(filterextra) > 0:
                    infostr2 = ut.repr2(filterextra, nl=False, explicit=False)
                    logger.info('[%s]      %s' % (self.prefix.upper(), infostr2))

        def __enter__(self):
            aids = ut.get_var_from_stack('avail_aids', verbose=False)
            self.num_before = len(aids)

        def __exit__(self, exc_type, exc_value, exc_traceback):
            if verbose:
                aids = ut.get_var_from_stack('avail_aids', verbose=False)
                num_after = len(aids)
                num_removed = self.num_before - num_after
                if num_removed > 0 or verbose > 1:
                    logger.info(
                        '[%s]   ... removed %d annots. %d remain'
                        % (self.prefix.upper(), num_removed, num_after)
                    )

        @staticmethod
        def report_annot_stats(ibs, aids, prefix, name_suffix, statskw={}):
            if verbose > 1:
                with ut.Indenter('[%s]  ' % (prefix.upper(),)):
                    # TODO: helpx on statskw
                    # statskw = dict(per_name_vpedge=None, per_name=None)
                    dict_name = prefix + 'aid_stats' + name_suffix
                    # hashid, per_name, per_qual, per_vp, per_name_vpedge,
                    # per_image, min_name_hourdist
                    ibs.print_annot_stats(aids, prefix=prefix, label=dict_name, **statskw)

        # def report_annotconfig_stats(ref_aids, aids):
        #    with ut.Indenter('  '):
        #        ibs.print_annotconfig_stats(ref_aids, avail_aids)

        @staticmethod
        def startfilter(withpre=True):
            """
            Args:
                withpre (bool): if True reports stats before filtering
            """
            if verbose:
                prefix = ut.get_var_from_stack('prefix', verbose=False)
                logger.info('[%s] * [%s] %sAIDS' % (prefix.upper(), filtertype, prefix))
                if verbose > 1 and withpre:
                    ibs = ut.get_var_from_stack('ibs', verbose=False)
                    aids = ut.get_var_from_stack('avail_aids', verbose=False)
                    VerbosityContext.report_annot_stats(ibs, aids, prefix, '_pre')

        @staticmethod
        def endfilter(withpost=True):
            if verbose:
                ibs = ut.get_var_from_stack('ibs', verbose=False)
                aids = ut.get_var_from_stack('avail_aids', verbose=False)
                prefix = ut.get_var_from_stack('prefix', verbose=False)
                hashid = ibs.get_annot_hashid_semantic_uuid(aids, prefix=prefix.upper())
                if withpost:
                    if verbose > 1:
                        VerbosityContext.report_annot_stats(ibs, aids, prefix, '_post')
                logger.info('[%s] * HAHID: %s' % (prefix.upper(), hashid))
                logger.info(
                    '[%s] * [%s]: len(avail_%saids) = %r\n'
                    % (prefix.upper(), filtertype, prefix, len(aids))
                )

    return VerbosityContext



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import cv2
import os
import random

# import xml.etree.ElementTree as xml

__location__ = os.path.realpath(os.path.join(os.getcwd(), os.path.dirname(__file__)))


def _kwargs(kwargs, key, value):
    if key not in kwargs.keys():
        kwargs[key] = value


def get(et, category, text=True, singularize=True):
    temp = [(_object.text if text else _object) for _object in et.findall(category)]
    if len(temp) == 1 and singularize:
        temp = temp[0]
    return temp


def histogram(_list):
    retDict = {}
    for value in _list:
        if value in retDict:
            retDict[value] += 1
        else:
            retDict[value] = 1
    return retDict


def openImage(filename, color=False, alpha=False):
    if not os.path.exists(filename):
        return None

    if not color:
        mode = 0  # Greyscale by default
    elif not alpha:
        mode = 1  # Color without alpha channel
    else:
        mode = -1  # Color with alpha channel

    return cv2.imread(filename, mode)


def randInt(lower, upper):
    return random.randint(lower, upper)


def randColor():
    return [randInt(50, 205), randInt(50, 205), randInt(50, 205)]



# -*- coding: utf-8 -*-
import logging
import torch
import torchvision
import utool as ut

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


class Siamese(torch.nn.Module):
    """
    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.algo.verif.siamese import *
        >>> self = Siamese()
    """

    def __init__(self):
        ut.super2(Siamese, self).__init__()
        self.resnet = torchvision.models.resnet50(pretrained=True)
        self.num_fcin = self.resnet.fc.in_features
        # replace the last layer of resnet
        self.resnet.fc = torch.nn.Linear(self.num_fcin, 500)
        self.pdist = torch.nn.PairwiseDistance(p=2)

    def forward(self, input1, input2):
        """
        Compute a resnet50 vector for each input and look at the L2 distance
        between the vectors.
        """
        output1 = self.resnet(input1)
        output2 = self.resnet(input2)
        output = self.pdist(output1, output2)

        return output


def visualize():
    import networkx as nx
    import torch
    from torch.autograd import Variable

    def make_nx(var, params):
        param_map = {id(v): k for k, v in params.items()}
        logger.info(param_map)
        node_attr = dict(
            style='filled',
            shape='box',
            align='left',
            fontsize='12',
            ranksep='0.1',
            height='0.2',
        )
        seen = set()
        G = nx.DiGraph()

        def size_to_str(size):
            return '(' + (', ').join(['%d' % v for v in size]) + ')'

        def build_graph(var):
            if var not in seen:
                if torch.is_tensor(var):
                    G.add_node(
                        id(var),
                        label=size_to_str(var.size()),
                        fillcolor='orange',
                        **node_attr,
                    )
                elif hasattr(var, 'variable'):
                    u = var.variable
                    node_name = '%s\n %s' % (param_map.get(id(u)), size_to_str(u.size()))
                    G.add_node(
                        id(var), label=node_name, fillcolor='lightblue', **node_attr
                    )
                else:
                    G.add_node(id(var), label=str(type(var).__name__), **node_attr)
                seen.add(var)
                if hasattr(var, 'next_functions'):
                    for u in var.next_functions:
                        if u[0] is not None:
                            G.add_edge(id(u[0]), id(var))
                            build_graph(u[0])
                if hasattr(var, 'saved_tensors'):
                    for t in var.saved_tensors:
                        G.add_edge(id(t), id(var))
                        build_graph(t)

        build_graph(var.grad_fn)
        return G

    # inputs = torch.randn(1, 3, 224, 224)
    # resnet18 = models.resnet18()
    # y = resnet18(Variable(inputs))

    inputs = torch.randn(1, 3, 224, 224)
    # model = torchvision.models.resnet18()
    model = torchvision.models.resnet50()

    model = Siamese()

    # y = model(Variable(inputs))
    y = model(Variable(inputs), Variable(inputs))

    params = model.state_dict()
    G = make_nx(y, params)

    import wbia.plottool as pt

    pt.dump_nx_ondisk(G, './pytorch_network.png')
    ut.startfile('./pytorch_network.png')
    # pt.show_nx(G, arrow_width=1)
    # pt.zoom_factory()
    # pt.pan_factory()



# -*- coding: utf-8 -*-
"""
Dependencies: flask, tornado
"""
import logging
from flask import current_app
from wbia.control import controller_inject
from wbia.web import appfuncs as appf
import utool as ut
from wbia.web import routes

(print, rrr, profile) = ut.inject2(__name__)
logger = logging.getLogger('wbia')

register_route = controller_inject.get_wbia_flask_route(__name__)


def get_associations_dict(ibs, desired_species=None, **kwargs):
    import itertools

    if ibs.dbname == 'ZEBRA_Kaia':
        valid_aid_set = set(ibs.get_valid_aids())
        imageset_list = ibs._princeton_kaia_imageset_filtering(**kwargs)
    else:
        valid_aid_set = set(ibs.get_valid_aids())
        imageset_list = ibs.get_valid_imgsetids(is_special=False)

    valid_nid_set = ibs.get_annot_nids(valid_aid_set)
    valid_nid_set = set([nid for nid in valid_nid_set if nid > 0])

    imageset_text_list = ibs.get_imageset_text(imageset_list)
    time_list = ibs.get_imageset_start_time_posix(imageset_list)
    nids_list = ibs.get_imageset_nids(imageset_list)

    ibs.delete_empty_nids()

    def _associate(dict_, name1, name2, label):
        if name2 < name1:
            name1, name2 = name2, name1
        if name1 not in dict_:
            dict_[name1] = {}
        if name2 not in dict_[name1]:
            dict_[name1][name2] = []
        dict_[name1][name2].append('%s' % (label,))

    assoc_dict = {}
    for imageset_text, time_, nid_list in zip(imageset_text_list, time_list, nids_list):
        if desired_species is not None:
            aids_list = ibs.get_name_aids(nid_list)

            flag_list = []
            for nid, aid_list in zip(nid_list, aids_list):
                aid_list = list(set(aid_list) & set(valid_aid_set))
                if len(aid_list) == 0:
                    flag = False
                else:
                    species_list = ibs.get_annot_species(aid_list)
                    species = max(set(species_list), key=species_list.count)
                    if ibs.dbname == 'ZEBRA_Kaia' and desired_species == 'zebra':
                        flag = species in ['zebra_plains', 'zebra_grevys']
                    else:
                        flag = species == desired_species
                flag_list.append(flag)
            nid_list = ut.compress(nid_list, flag_list)

        nid_list = list(set(nid_list) & valid_nid_set)
        name_list = ibs.get_name_texts(nid_list)
        # Add singles
        for name in name_list:
            _associate(assoc_dict, name, name, imageset_text)
            # _associate(assoc_dict, name, name, time_)
        # Add pairs
        comb_list = itertools.combinations(name_list, 2)
        for name1, name2 in sorted(list(comb_list)):
            _associate(assoc_dict, name1, name2, imageset_text)
            # _associate(assoc_dict, name, name, time_)

    return assoc_dict


@register_route('/csv/princeton/associations/list/', methods=['GET'])
def download_associations_list(**kwargs):
    ibs = current_app.ibs

    key_str_list = []
    for key in sorted(kwargs.keys()):
        key_str = '%s=%s' % (key, kwargs[key])
        key_str_list.append(key_str)
    key_str = '.'.join(key_str_list)
    if len(key_str) > 0:
        key_str += '.'

    filename = 'associations.list.%scsv' % (key_str,)
    assoc_dict = get_associations_dict(ibs, **kwargs)

    combined_list = []
    max_length = 0
    for name1 in sorted(assoc_dict.keys()):
        for name2 in sorted(assoc_dict[name1].keys()):
            id_list = sorted(set(assoc_dict[name1][name2]))
            id_list = [id_.replace(',', ':COMMA:') for id_ in id_list]
            max_length = max(max_length, len(id_list))
            args = (
                name1,
                name2,
                len(id_list),
                ','.join(id_list),
            )
            combined_str = '%s,%s,%s,%s' % args
            combined_list.append(combined_str)

    if max_length == 1:
        # name_header_str = 'TIME'
        name_header_str = 'ENCOUTNER'
    else:
        # name_header_str = ','.join([ 'TIME%d' % (i + 1, ) for i in range(max_length) ])
        name_header_str = ','.join(['ENCOUNTER%d' % (i + 1,) for i in range(max_length)])
    combined_str = '\n'.join(combined_list)
    combined_str = 'NAME1,NAME2,ASSOCIATIONS,%s\n' % (name_header_str,) + combined_str
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/princeton/associations/matrix/', methods=['GET'])
def download_associations_matrix(**kwargs):
    ibs = current_app.ibs

    key_str_list = []
    for key in sorted(kwargs.keys()):
        key_str = '%s=%s' % (key, kwargs[key])
        key_str_list.append(key_str)
    key_str = '.'.join(key_str_list)
    if len(key_str) > 0:
        key_str += '.'

    filename = 'associations.matrix.%scsv' % (key_str,)

    assoc_dict = get_associations_dict(ibs, **kwargs)
    assoc_list = sorted(assoc_dict.keys())
    # max_length = len(assoc_list)

    combined_list = []
    for index1, name1 in enumerate(assoc_list):
        temp_list = [name1]
        for index2, name2 in enumerate(assoc_list):
            value = assoc_dict[name1].get(name2, [])
            value_len = len(value)
            value_str = '' if value_len == 0 else value_len
            temp_list.append('%s' % (value_str,))
        temp_str = ','.join(temp_list)
        combined_list.append(temp_str)

    # if max_length == 1:
    #     name_header_str = 'NAME'
    # else:
    #     name_header_str = ','.join([ 'NAME%d' % (i + 1, ) for i in range(max_length) ])
    name_header_str = ','.join(assoc_list)
    combined_str = '\n'.join(combined_list)
    combined_str = 'MATRIX,%s\n' % (name_header_str,) + combined_str
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/princeton/sightings/', methods=['GET'])
def download_sightings(**kwargs):
    ibs = current_app.ibs
    filename = 'sightings.csv'
    kaia = ibs.dbname == 'ZEBRA_Kaia'
    sightings = routes.sightings(html_encode=False, kaia=kaia, **kwargs)
    return appf.send_csv_file(sightings, filename)


@register_route('/csv/princeton/images/', methods=['GET'])
def get_image_info(**kwargs):
    import datetime

    ibs = current_app.ibs
    filename = 'images.csv'

    valid_aid_set = set(ibs.get_valid_aids())

    gid_list = sorted(ibs.get_valid_gids())
    gname_list = ibs.get_image_gnames(gid_list)
    datetime_list = ibs.get_image_unixtime(gid_list)
    datetime_list_ = [
        datetime.datetime.fromtimestamp(datetime_).strftime('%Y-%m-%d %H:%M:%S')
        for datetime_ in datetime_list
    ]
    lat_list = ibs.get_image_lat(gid_list)
    lon_list = ibs.get_image_lon(gid_list)
    note_list = ibs.get_image_notes(gid_list)
    party_list = []
    contributor_list = []
    for note in note_list:
        try:
            note = note.split(',')
            party, contributor = note[:2]
            party_list.append(party)
            contributor_list.append(contributor)
        except Exception:
            party_list.append('UNKNOWN')
            contributor_list.append('UNKNOWN')

    zipped_list = zip(
        gid_list,
        gname_list,
        datetime_list_,
        lat_list,
        lon_list,
        party_list,
        contributor_list,
        note_list,
    )
    aids_list = ibs.get_image_aids(gid_list)
    aids_list = [
        [aid_ for aid_ in aid_list_ if aid_ in valid_aid_set] for aid_list_ in aids_list
    ]
    names_list = [ibs.get_annot_name_texts(aid_list) for aid_list in aids_list]
    combined_list = [
        ','.join(map(str, list(zipped) + name_list))
        for zipped, aid_list, name_list in zip(zipped_list, aids_list, names_list)
        if ibs.dbdir != 'ZEBRA_Kaia' or len(aid_list) > 0
    ]
    max_length = 0
    for name_list in names_list:
        max_length = max(max_length, len(name_list))
    if max_length == 1:
        name_header_str = 'NAME'
    else:
        name_header_str = ','.join(['NAME%d' % (i + 1,) for i in range(max_length)])
    combined_str = '\n'.join(combined_list)
    combined_str = (
        'GID,FILENAME,TIMESTAMP,GPSLAT,GPSLON,PARTY,CONTRIBUTOR,NOTES,%s\n'
        % (name_header_str,)
        + combined_str
    )
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/princeton/demographics/', methods=['GET'])
def get_demographic_info(**kwargs):
    ibs = current_app.ibs
    filename = 'demographics.csv'
    nid_list = sorted(ibs.get_valid_nids())
    name_list = ibs.get_name_texts(nid_list)
    sex_list = ibs.get_name_sex_text(nid_list)
    min_ages_list = ibs.get_name_age_months_est_min(nid_list)
    max_ages_list = ibs.get_name_age_months_est_max(nid_list)

    def _get_primary_species(aid_list):
        if len(aid_list) == 0:
            species = None
        else:
            species_list = ibs.get_annot_species_texts(aid_list)
            species = max(set(species_list), key=species_list.count)
        return species

    age_list = []
    for min_ages, max_ages in zip(min_ages_list, max_ages_list):
        if len(set(min_ages)) > 1 or len(set(max_ages)) > 1:
            age_list.append('AMBIGUOUS')
            continue
        min_age = None
        max_age = None
        if len(min_ages) > 0:
            min_age = min_ages[0]
        if len(max_ages) > 0:
            max_age = max_ages[0]
        # Histogram
        if (min_age is None and max_age is None) or (min_age is -1 and max_age is -1):
            age_list.append('UNREVIEWED')
            continue
        # Bins
        if min_age is None and max_age == 2:
            age = '0-3 Months'
        elif min_age == 3 and max_age == 5:
            age = '3-6 Months'
        elif min_age == 6 and max_age == 11:
            age = '6-12 Months'
        elif min_age == 12 and max_age == 23:
            age = 'Yearling'
        elif min_age == 24 and max_age == 35:
            age = '2-Year-Old'
        elif min_age == 36 and max_age is None:
            age = 'Adult'
        elif min_age is None and max_age is None:
            age = 'Unknown'
        else:
            age = 'Unknown'
        age_list.append(age)

    zipped_list = zip(nid_list, name_list, sex_list, age_list)
    combined_list = [','.join(map(str, list(zipped))) for zipped in zipped_list]
    combined_str = '\n'.join(combined_list)
    combined_str = 'NID,NAME,SEX,AGE\n' + combined_str
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/princeton/special/kaia-dung-samples/', methods=['GET'])
def get_annotation_special_kaia_dung_samples(**kwargs):
    ibs = current_app.ibs

    filename = 'dungsamples.csv'
    aid_list = ibs.get_valid_aids()
    dungsample_list = ibs.get_annot_notes(aid_list)
    flag_list = [len(dungsample) > 0 for dungsample in dungsample_list]
    aid_list = ut.compress(aid_list, flag_list)
    dungsample_list = ut.compress(dungsample_list, flag_list)
    nid_list = ibs.get_annot_nids(aid_list)
    name_list = ibs.get_annot_name_texts(aid_list)

    assoc_dict = get_associations_dict(ibs, desired_species='zebra', tier=1)
    encounter_str_list = []
    max_length = 0
    for name in sorted(name_list):
        id_list = sorted(set(assoc_dict[name][name]))
        id_list = [id_.replace(',', ':COMMA:') for id_ in id_list]
        max_length = max(max_length, len(id_list))
        encounter_str = ','.join(id_list)
        encounter_str_list.append(encounter_str)

    sex_list = ibs.get_annot_sex(aid_list)
    age_min_list = ibs.get_annot_age_months_est_min(aid_list)
    age_max_list = ibs.get_annot_age_months_est_max(aid_list)
    condition_list = ibs.get_annot_qualities(aid_list)

    sex_list_ = []
    age_list_ = []
    condition_list_ = []
    zipped = zip(sex_list, age_min_list, age_max_list, condition_list)
    for sex_, age_min, age_max, condition_ in zipped:
        if sex_ == 1:
            sex_ = 'male'
        elif sex_ == 0:
            sex_ = 'female'
        else:
            sex_ = 'unknown'

        if age_min is None and age_max == 2:
            age_ = 'age1'
        elif age_min == 3 and age_max == 5:
            age_ = 'age2'
        elif age_min == 6 and age_max == 11:
            age_ = 'age3'
        elif age_min == 12 and age_max == 23:
            age_ = 'age4'
        elif age_min == 24 and age_max == 35:
            age_ = 'age5'
        elif age_min == 36 and age_max is None:
            age_ = 'age6'
        elif age_min is None and age_max is None:
            age_ = 'unknown'
        else:
            age_ = 'unknown'

        if condition_ is None:
            condition_ = 0

        assert age_ in ['age1', 'age2', 'age3', 'age4', 'age5', 'age6', 'unknown']
        assert sex_ in ['male', 'female', 'unknown']
        assert 0 <= condition_ and condition_ <= 5

        if sex_ == 'male':
            sex_ = 'Male'
        elif sex_ == 'female':
            sex_ = 'Female'
        elif sex_ == 'unknown':
            sex_ = 'Unknown'
        else:
            sex_ = 'Unknown'

        if age_ == 'age1':
            age_ = '0-3 Months'
        elif age_ == 'age2':
            age_ = '3-6 Months'
        elif age_ == 'age3':
            age_ = '6-12 Months'
        elif age_ == 'age4':
            age_ = 'Yearling'
        elif age_ == 'age5':
            age_ = '2-Year-Old'
        elif age_ == 'age6':
            age_ = 'Adult'
        elif age_ == 'unknown':
            age_ = 'Unknown'
        else:
            age_ = 'Unknown'

        sex_list_.append(sex_)
        age_list_.append(age_)
        condition_list_.append(condition_)

    if max_length == 1:
        name_header_str = 'ENCOUTNER'
    else:
        name_header_str = ','.join(['ENCOUNTER%d' % (i + 1,) for i in range(max_length)])

    zipped_list = sorted(
        zip(
            name_list,
            nid_list,
            aid_list,
            dungsample_list,
            age_list_,
            sex_list_,
            condition_list_,
            encounter_str_list,
        )
    )
    combined_list = [','.join(map(str, list(zipped_))) for zipped_ in zipped_list]
    combined_str = '\n'.join(combined_list)
    combined_str = (
        'NAME,NID,AID,DUNGSAMPLE,AGE,SEX,CONDITION,%s\n' % (name_header_str,)
        + combined_str
    )
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/princeton/special/monica-laurel-max/', methods=['GET'])
def get_annotation_special_monica_laurel_max(desired_species=None, **kwargs):
    ibs = current_app.ibs
    filename = 'special.monica-laurel-max.csv'

    def _process_annot_name_uuids_dict(ibs, filepath):
        import uuid

        auuid_list = []
        nuuid_list = []
        with open(filepath, 'r') as file_:
            for line in file_.readlines():
                line = line.strip().split(',')
                auuid = uuid.UUID(line[0])
                nuuid = None if line[1] == 'None' else uuid.UUID(line[1])
                auuid_list.append(auuid)
                nuuid_list.append(nuuid)

        annot_rowid_list = ibs.get_annot_aids_from_uuid(auuid_list)
        name_rowid_list = ibs.get_name_rowids_from_uuid(nuuid_list)

        zipped = zip(annot_rowid_list, name_rowid_list)
        mapping_dict = {aid: nid for aid, nid in zipped if aid is not None}
        return mapping_dict

    aid_list = sorted(ibs.get_valid_aids())
    annot_uuid_list = ibs.get_annot_uuids(aid_list)
    logger.info('Found %d aids' % (len(aid_list),))
    nid_list = ibs.get_annot_nids(aid_list)
    name_uuid_list = ibs.get_name_uuids(nid_list)
    name_list = ibs.get_name_texts(nid_list)
    species_list = ibs.get_annot_species_texts(aid_list)
    sex_list = ibs.get_name_sex_text(nid_list)
    age_list = ibs.get_annot_age_months_est_texts(aid_list)
    gid_list = ibs.get_annot_gids(aid_list)
    contrib_list = ibs.get_image_contributor_tag(gid_list)
    gname_list = ibs.get_image_gnames(gid_list)
    imageset_rowids_list = ibs.get_image_imgsetids(gid_list)
    imageset_rowids_set = map(set, imageset_rowids_list)

    special_imageset_rowid_set = set(ibs.get_valid_imgsetids(is_special=True))
    imagesets_list = [
        list(imageset_rowid_set - special_imageset_rowid_set)
        for imageset_rowid_set in imageset_rowids_set
    ]

    imageset_list = [_[0] if len(_) > 0 else None for _ in imagesets_list]
    imageset_text_list = ibs.get_imageset_text(imageset_list)
    imageset_metadata_list = ibs.get_imageset_metadata(imageset_list)
    annot_metadata_list = ibs.get_annot_metadata(aid_list)

    assert len(imageset_metadata_list) == len(annot_metadata_list)

    imageset_metadata_list_ = ibs.get_imageset_metadata(ibs.get_valid_imgsetids())
    imageset_metadata_key_list = sorted(
        set(
            ut.flatten(
                [
                    imageset_metadata_dict_.keys()
                    for imageset_metadata_dict_ in imageset_metadata_list_
                ]
            )
        )
    )
    imageset_metadata_key_str = ','.join(imageset_metadata_key_list)

    annot_metadata_list_ = ibs.get_annot_metadata(ibs.get_valid_aids())
    annot_metadata_key_list = sorted(
        set(
            ut.flatten(
                [
                    annot_metadata_dict_.keys()
                    for annot_metadata_dict_ in annot_metadata_list_
                ]
            )
        )
    )
    annot_metadata_key_str = ','.join(annot_metadata_key_list)

    if 'Monica-Laurel' in ibs.dbdir:
        import wbia

        ibs1 = wbia.opendb('/home/zebra/Desktop/Monica/', web=False)
        ibs2 = wbia.opendb('/home/zebra/Desktop/Laurel/', web=False)
    if 'Monica-Max' in ibs.dbdir:
        import wbia

        ibs1 = wbia.opendb('/home/zebra/Desktop/Monica/', web=False)
        ibs2 = wbia.opendb('/home/zebra/Desktop/Max/', web=False)
    else:
        ibs1 = None
        ibs2 = None

    line_list = []
    zipped = zip(
        nid_list,
        aid_list,
        name_uuid_list,
        annot_uuid_list,
        name_list,
        species_list,
        sex_list,
        age_list,
        gname_list,
        contrib_list,
        imageset_list,
        imageset_text_list,
        imageset_metadata_list,
        annot_metadata_list,
    )
    zipped = sorted(list(zipped))
    for args in zipped:
        (
            nid,
            aid,
            name_uuid,
            annot_uuid,
            name,
            species,
            sex,
            age,
            gname,
            contrib,
            imageset_rowid,
            imageset_text,
            imageset_metadata_dict,
            annot_metadata_dict,
        ) = args

        contrib_str = '' if contrib is None else contrib.split(',')[0].upper()

        if desired_species is not None and species != desired_species:
            continue

        if nid <= 0:
            continue

        nid_old = ''
        name_old = ''
        name_changed = False
        cross_database_match = False

        try:
            if ibs1 is not None and ibs2 is not None:
                aid1 = ibs1.get_annot_aids_from_uuid(annot_uuid)
                aid2 = ibs2.get_annot_aids_from_uuid(annot_uuid)

                if aid1 is not None:
                    assert aid2 is None
                    name_uuid_old = ibs1.get_annot_name_uuids(aid1)
                elif aid2 is not None:
                    assert aid1 is None
                    name_uuid_old = ibs2.get_annot_name_uuids(aid2)
                else:
                    raise AssertionError('Should be in one of these original databases')

                if name_uuid_old != name_uuid:
                    name_changed = True
                    if name_uuid_old is None:
                        nid_old = 'UNKNOWN NID'
                        name_old = 'UNKNOWN NAME'
                    else:
                        nid_old = ibs.get_name_rowids_from_uuid(name_uuid_old)
                        name_old = ibs.get_name_texts(nid_old)

                    cross_database_match = not name.startswith(contrib_str)
        except Exception:
            logger.info('ERROR WITH ABOVE')
            ut.embed()

        line_list_ = (
            [
                contrib_str,
                annot_uuid,
                aid,
                annot_uuid,
                nid,
                name,
                'Yes' if name_changed else '',
                'Yes' if cross_database_match else '',
                nid_old,
                name_old,
                species,
                sex,
                age,
                gname,
                imageset_rowid,
                imageset_text,
                '|',
            ]
            + [
                imageset_metadata_dict.get(imageset_metadata_key, '')
                for imageset_metadata_key in imageset_metadata_key_list
            ]
            + ['|']
            + [
                annot_metadata_dict.get(annot_metadata_key, '')
                for annot_metadata_key in annot_metadata_key_list
            ]
        )
        line_list_ = ['' if item is None else item for item in line_list_]
        line = ','.join(map(str, line_list_))
        line_list.append(line)

    combined_str = '\n'.join(line_list)
    combined_str = (
        'DB,Annotation UUID,AID,NID,Name,Name Changed,Cross-Database Match,Old NID,Old Name,Species,Sex,Age,Image Name,Encounter ID,Encounter Name,| SEPERATOR |,%s,| SEPERATOR |,%s\n'
        % (imageset_metadata_key_str, annot_metadata_key_str)
        + combined_str
    )
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/princeton/special/megan/', methods=['GET'])
def get_annotation_special_megan(**kwargs):
    ibs = current_app.ibs
    filename = 'special.megan.csv'

    gid_list = ibs.get_valid_gids()
    uri_list = ibs.get_image_uris(gid_list)
    path_list = ibs.get_image_uris_original(gid_list)

    tag_list = []
    date1_list = []
    date2_list = []
    note_list = []
    for path, gid in zip(path_list, gid_list):
        path = path.replace('/home/zebra/Desktop/MEGAN/', '')
        path = path.split('/')
        base = path[0]
        if base == 'SIDE1_3-22-2016-5-11-2016':  # HACK
            base = 'SIDE1_3-22-2016_5-11-2016'
        path = base.split('_')
        try:
            assert len(path) >= 3
            if len(path) > 3:
                note = path[3:]
                note = '_'.join(note)
                path = path[:3]
            else:
                note = ''
            assert len(path) == 3
            tag, date1, date2 = path
            date1 = date1.replace('-', '/')
            date2 = date2.replace('-', '/')
            tag_list.append(tag)
            date1_list.append(date1)
            date2_list.append(date2)
            note_list.append(note)
        except AssertionError:
            logger.info('ERROR PROCESSING: %r' % (base,))

    zipped = zip(gid_list, uri_list, tag_list, date1_list, date2_list, note_list)
    combined_list = [','.join(map(str, value_list)) for value_list in zipped]
    combined_str = '\n'.join(combined_list)
    combined_str = 'GID,FILENAME,LOCATION,DATE1,DATE2,NOTE\n' + combined_str
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/nids_with_gids/', methods=['GET'])
def get_nid_with_gids_csv(**kwargs):
    ibs = current_app.ibs
    filename = 'nids_with_gids.csv'
    combined_dict = ibs.get_name_nids_with_gids()
    combined_list = [
        ','.join(map(str, [nid] + [name] + gid_list))
        for name, (nid, gid_list) in sorted(list(combined_dict.iteritems()))
    ]
    combined_str = '\n'.join(combined_list)
    max_length = 0
    for aid_list in combined_dict.values(**kwargs):
        max_length = max(max_length, len(aid_list[1]))
    if max_length == 1:
        gid_header_str = 'GID'
    else:
        gid_header_str = ','.join(['GID%d' % (i + 1,) for i in range(max_length)])
    combined_str = 'NID,NAME,%s\n' % (gid_header_str,) + combined_str
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/gids_with_aids/', methods=['GET'])
def get_gid_with_aids_csv(**kwargs):
    ibs = current_app.ibs
    combined_dict = ibs.get_image_gids_with_aids()
    filename = 'gids_with_aids.csv'
    combined_list = [
        ','.join(map(str, [gid] + aid_list))
        for gid, aid_list in sorted(list(combined_dict.iteritems()))
    ]
    combined_str = '\n'.join(combined_list)
    max_length = 0
    for aid_list in combined_dict.values(**kwargs):
        max_length = max(max_length, len(aid_list))
    if max_length == 1:
        aid_header_str = 'AID'
    else:
        aid_header_str = ','.join(['AID%d' % (i + 1,) for i in range(max_length)])
    combined_str = 'GID,%s\n' % (aid_header_str,) + combined_str
    return appf.send_csv_file(combined_str, filename)


@register_route('/csv/image/', methods=['GET'])
def get_gid_list_csv(**kwargs):
    filename = 'gids.csv'
    ibs = current_app.ibs
    gid_list = ibs.get_valid_gids()
    return_str = '\n'.join(map(str, gid_list))
    return_str = 'GID\n' + return_str
    return appf.send_csv_file(return_str, filename)


@register_route('/csv/annot/', methods=['GET'])
def get_aid_list_csv(**kwargs):
    filename = 'aids.csv'
    ibs = current_app.ibs
    aid_list = ibs.get_valid_aids()
    return_str = '\n'.join(map(str, aid_list))
    return_str = 'AID\n' + return_str
    return appf.send_csv_file(return_str, filename)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from wbia.plottool import viz_image2
from wbia.plottool import draw_func2 as df2
from wbia.plottool import plot_helpers as ph
import utool
import numpy as np
from wbia.plottool.tests.test_helpers import dummy_bbox, imread_many


def _test_viz_image(imgpaths):
    nImgs = len(imgpaths)
    assert len(imgpaths) < 20, '%d > 20 out of scope of this test' % nImgs
    tau = np.pi * 2
    fnum = 1
    img_list = imread_many(imgpaths)
    nRows, nCols = ph.get_square_row_cols(nImgs)
    print('[viz*] r=%r, c=%r' % (nRows, nCols))
    # gs2 = gridspec.GridSpec(nRows, nCols)
    pnum_ = df2.get_pnum_func(nRows, nCols)

    fig = df2.figure(fnum=fnum, pnum=pnum_(0))
    fig.clf()
    for px, img in enumerate(img_list):
        title = 'test title'
        bbox_list = [dummy_bbox(img), dummy_bbox(img, (-0.25, -0.25), 0.1)]
        theta_list = [tau * 0.7, tau * 0.9]
        sel_list = [True, False]
        label_list = ['test label', 'lbl2']
        viz_image2.show_image(
            img,
            bbox_list=bbox_list,
            title=title,
            sel_list=sel_list,
            label_list=label_list,
            theta_list=theta_list,
            fnum=fnum,
            pnum=pnum_(px),
        )


if __name__ == '__main__':
    TEST_IMAGES_URL = 'https://wildbookiarepository.azureedge.net/data/testdata.zip'
    test_image_dir = utool.grab_zipped_url(TEST_IMAGES_URL, appname='utool')
    imgpaths = utool.list_images(test_image_dir, fullpath=True)  # test image paths
    _test_viz_image(imgpaths)
    exec(df2.present())



# -*- coding: utf-8 -*-
"""
Small GUI for asking the user to enter the clock time shown, and moving along a gid list if the first image isn't a clock
"""
import logging
from functools import partial
from six.moves import range
from time import mktime
from datetime import date, datetime
import utool as ut
import wbia.guitool as gt
from wbia.guitool.__PYQT__ import QtWidgets
from wbia.guitool.__PYQT__.QtCore import Qt
import wbia.plottool as pt

(print, rrr, profile) = ut.inject2(__name__, '[co_gui]')
logger = logging.getLogger('wbia')


class ClockOffsetWidget(QtWidgets.QWidget):
    def __init__(co_wgt, ibs, gid_list, parent=None, hack=False):
        logger.info('[co_gui] Initializing')
        logger.info('[co_gui] gid_list = %r' % (gid_list,))

        QtWidgets.QWidget.__init__(co_wgt, parent=parent)

        co_wgt.fnum = pt.next_fnum()

        co_wgt.main_layout = QtWidgets.QVBoxLayout(co_wgt)

        co_wgt.text_layout = gt.newWidget(
            co_wgt, orientation=Qt.Vertical, verticalStretch=10
        )
        co_wgt.main_layout.addWidget(co_wgt.text_layout)

        co_wgt.control_layout = gt.newWidget(
            co_wgt,
            orientation=Qt.Vertical,
            verticalSizePolicy=QtWidgets.QSizePolicy.MinimumExpanding,
        )
        co_wgt.main_layout.addWidget(co_wgt.control_layout)

        co_wgt.button_layout = gt.newWidget(co_wgt, orientation=Qt.Horizontal)
        co_wgt.control_layout.addWidget(co_wgt.button_layout)

        co_wgt.combo_layout = gt.newWidget(co_wgt, orientation=Qt.Horizontal)
        co_wgt.control_layout.addWidget(co_wgt.combo_layout)

        co_wgt.hack = hack  # hack for edting the time of a single image

        co_wgt.imfig = None
        co_wgt.imax = None

        co_wgt.dtime = None

        co_wgt.ibs = ibs
        co_wgt.gid_list = gid_list
        co_wgt.current_gindex = 0
        co_wgt.offset = 0
        # Set image datetime with first image
        co_wgt.get_image_datetime_()
        co_wgt.add_label()
        co_wgt.add_combo_boxes()
        co_wgt.add_buttons()
        co_wgt.update_ui()

    def resizeEvent(co_wgt, event):
        super(ClockOffsetWidget, co_wgt).resizeEvent(event)
        if co_wgt.image_label is not None:
            # Hack use signals and slots
            if hasattr(co_wgt.image_label, '_on_resize_slot'):
                co_wgt.image_label._on_resize_slot()
        # logger.info('resizeEvent')

    def get_image_datetime_(co_wgt):
        # Function that extracts the unixtime from an image and stores it
        utime = co_wgt.ibs.get_image_unixtime(co_wgt.gid_list[co_wgt.current_gindex])
        if not isinstance(utime, (float, int)) or utime is None:
            utime = -1
        utime = float(utime)
        co_wgt.dtime = datetime.fromtimestamp(utime)

    def update_ui(co_wgt):
        if not co_wgt.hack:
            if co_wgt.current_gindex == 0:
                co_wgt.button_list[0].setEnabled(False)
            else:
                co_wgt.button_list[0].setEnabled(True)
            if co_wgt.current_gindex == len(co_wgt.gid_list) - 1:
                co_wgt.button_list[1].setEnabled(False)
            else:
                co_wgt.button_list[1].setEnabled(True)

        def extract_tuple(li, idx):
            return list(zip(*li)[idx])

        # Update option setting, assume datetime has been updated
        co_wgt.combo_list[1].setCurrentIndex(
            extract_tuple(co_wgt.opt_list['year'], 1).index(co_wgt.dtime.year)
        )
        co_wgt.combo_list[3].setCurrentIndex(
            extract_tuple(co_wgt.opt_list['month'], 1).index(co_wgt.dtime.month)
        )
        co_wgt.combo_list[5].setCurrentIndex(
            extract_tuple(co_wgt.opt_list['day'], 1).index(co_wgt.dtime.day)
        )
        co_wgt.combo_list[7].setCurrentIndex(
            extract_tuple(co_wgt.opt_list['hour'], 1).index(co_wgt.dtime.hour)
        )
        co_wgt.combo_list[9].setCurrentIndex(
            extract_tuple(co_wgt.opt_list['minute'], 1).index(co_wgt.dtime.minute)
        )
        co_wgt.combo_list[11].setCurrentIndex(
            extract_tuple(co_wgt.opt_list['second'], 1).index(co_wgt.dtime.second)
        )

        # Redraw image
        if not co_wgt.hack:
            if co_wgt.imfig is not None:
                pt.close_figure(co_wgt.imfig)
            image = co_wgt.ibs.get_images(co_wgt.gid_list[co_wgt.current_gindex])
            figtitle = 'Time Synchronization Picture'
            co_wgt.imfig, co_wgt.imax = pt.imshow(image, fnum=co_wgt.fnum, title=figtitle)
            co_wgt.imfig.show()

    def show_helpmsg(co_wgt):
        msg = ut.textblock(
            """
            This step is for synchronizing the time of the images being imported
            with the actual time of the imageset.

            Use the Previous and Next buttons until the 'Time Synchronization Picture' is of the clock
            taken at the beginning of the imageset.

            Once found, change the date and time in the boxes below to match the time of the clock in
            the image, and correct the date if necessary.

            Once done, click 'Set' and the difference will be applied to all images currently
            being imported.  If you are sure the camera was synchronized, you
            can skip this step by pressing 'Skip'.
            """
        )
        gt.user_info(co_wgt, msg=msg, title='Time Sync Help')

    def add_label(co_wgt):
        # Very simply adds the text
        _LABEL = partial(gt.newLabel, parent=co_wgt)
        if not co_wgt.hack:
            text = ut.codeblock(
                """
                * Find the image of the clock
                * Set the sliders to correspond with the clock
                * Click Set
                * Skip if time synchonization is not relevant to you
                """
            )
            main_label = _LABEL(text=text, align='left')
            co_wgt.text_layout.addWidget(main_label)
        else:
            text = ut.codeblock(
                """
                * Set the time for image %r
                """
                % (co_wgt.gid_list,)
            )
            gpath = co_wgt.ibs.get_image_paths(co_wgt.gid_list[co_wgt.current_gindex])
            image_label = _LABEL(text='', gpath=gpath)  # align='left')
            co_wgt.image_label = image_label
            co_wgt.text_layout.addWidget(image_label)
            # main_label = _LABEL(text=text, align='left')
            # co_wgt.text_layout.addWidget(main_label)

    def add_buttons(co_wgt):
        _BUTTON = partial(gt.newButton, parent=co_wgt)
        if not co_wgt.hack:
            co_wgt.button_list = [
                _BUTTON(text='Previous Image', clicked=co_wgt.go_prev),
                _BUTTON(text='Next Image', clicked=co_wgt.go_next),
                _BUTTON(text='Skip', clicked=co_wgt.cancel),
                _BUTTON(text='Set', clicked=co_wgt.accept),
                _BUTTON(text='Help', clicked=co_wgt.show_helpmsg),
            ]
        else:
            co_wgt.button_list = [
                _BUTTON(text='Accept', clicked=co_wgt.accept),
            ]
        # Copying from inspect_gui

        for button in co_wgt.button_list:
            co_wgt.button_layout.addWidget(button)

    def add_combo_boxes(co_wgt):
        _CBOX = partial(gt.newComboBox, parent=co_wgt)
        _LABEL = partial(gt.newLabel, parent=co_wgt, align='right')

        def to_opt_list(x):
            return [('%02d' % i, i) for i in x]

        year_list = [
            datetime.fromtimestamp(unixtime).year
            for unixtime in co_wgt.ibs.get_image_unixtime(co_wgt.gid_list)
            if unixtime is not None
        ]
        max_year = max(year_list + [date.today().year]) + 1
        min_year = min(year_list + [1970]) - 1
        # max_year = date.today().year + 1
        co_wgt.opt_list = {
            'year': to_opt_list(range(min_year, max_year)),
            'month': to_opt_list(range(1, 13)),
            'day': to_opt_list(
                range(1, 32)
            ),  # yes this will allow invalid dates, but it's not a real issue
            'hour': to_opt_list(range(24)),
            'minute': to_opt_list(range(60)),
            'second': to_opt_list(range(60)),
        }
        co_wgt.combo_list = [
            _LABEL(text='YYYY'),
            _CBOX(
                changed=partial(co_wgt.change_dt, 'year'),
                options=co_wgt.opt_list['year'],
                default=co_wgt.dtime.year,
            ),
            _LABEL(text='MM'),
            _CBOX(
                changed=partial(co_wgt.change_dt, 'month'),
                options=co_wgt.opt_list['month'],
                default=co_wgt.dtime.month,
            ),
            _LABEL(text='DD'),
            _CBOX(
                changed=partial(co_wgt.change_dt, 'day'),
                options=co_wgt.opt_list['day'],
                default=co_wgt.dtime.day,
            ),
            _LABEL(text='HH (24H)'),
            _CBOX(
                changed=partial(co_wgt.change_dt, 'hour'),
                options=co_wgt.opt_list['hour'],
                default=co_wgt.dtime.hour,
            ),
            _LABEL(text='mm'),
            _CBOX(
                changed=partial(co_wgt.change_dt, 'minute'),
                options=co_wgt.opt_list['minute'],
                default=co_wgt.dtime.minute,
            ),
            _LABEL(text='ss'),
            _CBOX(
                changed=partial(co_wgt.change_dt, 'second'),
                options=co_wgt.opt_list['second'],
                default=co_wgt.dtime.second,
            ),
        ]

        for combo in co_wgt.combo_list:
            co_wgt.combo_layout.addWidget(combo)

    @gt.slot_()
    def go_prev(co_wgt):
        # Decrement current gid index, then call update
        co_wgt.current_gindex -= 1
        co_wgt.get_image_datetime_()
        co_wgt.update_ui()

    @gt.slot_()
    def go_next(co_wgt):
        # Increment current gid index, then call update
        co_wgt.current_gindex += 1
        co_wgt.get_image_datetime_()
        co_wgt.update_ui()

    @gt.slot_()
    def cancel(co_wgt):
        # Just close
        pt.close_figure(co_wgt.imfig)
        co_wgt.close()

    @gt.slot_()
    def accept(co_wgt):
        # Calculate offset
        # Get current image's actual time
        image_time = co_wgt.ibs.get_image_unixtime(co_wgt.gid_list[co_wgt.current_gindex])
        # Get unixtime from current dtime
        input_time = mktime(co_wgt.dtime.timetuple())
        # Go through gid list, and add that offset to every unixtime
        offset = input_time - image_time
        logger.info('[co_gui] Unixtime offset is %d' % offset)
        # Set the unixtimes with the new ones
        utimes = co_wgt.ibs.get_image_unixtime(co_wgt.gid_list)
        new_utimes = [time + offset for time in utimes]
        # SHOULD WE BE USING image_timedelta_posix INSTEAD OF THIS?
        co_wgt.ibs.set_image_unixtime(co_wgt.gid_list, new_utimes)
        # Close
        if not co_wgt.hack:
            pt.close_figure(co_wgt.imfig)
        co_wgt.close()

    @gt.slot_(str, int, str)
    def change_dt(co_wgt, attr, val_ind, val):
        co_wgt.dtime = co_wgt.dtime.replace(**{attr: co_wgt.opt_list[attr][val_ind][1]})
        logger.info('[co_gui] Base datetime is %r' % co_wgt.dtime)


def clock_offset_test():
    r"""
    CommandLine:
        python -m wbia.gui.clock_offset_gui --test-clock_offset_test

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.gui.clock_offset_gui import *  # NOQA
        >>> # build test data
        >>> # execute function
        >>> result = clock_offset_test()
        >>> # verify results
        >>> print(result)
    """
    import wbia

    main_locals = wbia.main(db='testdb1')
    ibs = main_locals['ibs']
    gid_list = ibs.get_valid_gids()
    co = ClockOffsetWidget(ibs, gid_list, hack=True)
    co.show()
    wbia.main_loop(main_locals)
